
from functools import partial
from typing import Any, Callable, Optional, Union, MutableMapping
import six
from sqlalchemy import PickleType, Column
from sqlalchemy.orm import deferred
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import Mutable

import dill
import pyzstd

from ms_deisotope.data_source._compression import starts_with_zstd_magic


class LazyMutableMappingWrapper(Mutable, MutableMapping[str, Any]):
    _payload: Optional[bytes]
    _unpickler: Callable[[bytes], MutableMapping[str, Any]]
    _object: MutableMapping[str, Any]

    def __init__(self, payload=None, unpickler=None, _object=None):
        if payload is None and _object is None:
            _object = PartiallySerializedMutableMapping()
        if payload is not None and not starts_with_zstd_magic(payload):
            payload = pyzstd.compress(payload)
        self._payload = payload
        self._unpickler = unpickler
        self._object = _object

    def _load(self):
        if self._object is not None:
            return self._object
        self._object = self._unpickler.loads(pyzstd.decompress(self._payload))
        self._payload = None
        return self._object

    def keys(self):
        return self._load().keys()

    def values(self):
        return self._load().values()

    def items(self):
        return self._load().items()

    def __len__(self):
        return len(self._load())

    def __contains__(self, key):
        return key in self._load()

    def __getitem__(self, key):
        return self._load()[key]

    def __iter__(self):
        return iter(self._load())

    def __setitem__(self, key, value):
        """Detect dictionary set events and emit change events."""
        self._load()[key] = value
        self.changed()

    def setdefault(self, key, value):
        result = self._load().setdefault(key, value)
        self.changed()
        return result

    def __delitem__(self, key):
        """Detect dictionary del events and emit change events."""
        self._load().__delitem__(key)
        self.changed()

    def update(self, *a, **kw):
        self._load().update(*a, **kw)
        self.changed()

    def pop(self, *arg):
        result = self._load().pop(*arg)
        self.changed()
        return result

    def popitem(self):
        result = self._load().popitem()
        self.changed()
        return result

    def clear(self):
        self._load().clear()
        self.changed()

    def __getstate__(self):
        return self._payload, self._unpickler, self._object

    def __setstate__(self, state):
        try:
            (payload, unpickler, _object) = state
        except Exception:
            payload = None
            unpickler = None
            _object = state
        self._payload = payload
        self._unpickler = unpickler
        self._object = PartiallySerializedMutableMapping()
        self.update(_object or {})

    def unload(self):
        for k, v in self._load().store.items():
            if hasattr(v, 'serialized'):
                if v.serialized:
                    v.serialize()

    @classmethod
    def coerce(cls, key, value):
        """Convert plain dictionary to instance of this class."""
        if not isinstance(value, cls):
            if isinstance(value, dict):
                return cls(None, None, PartiallySerializedMutableMapping(value))
            elif isinstance(value, PartiallySerializedMutableMapping):
                return cls(None, None, value)
            elif isinstance(value, (str, bytes)):
                return cls(value, dill.loads)
            # We have an old `glycan_profiling` object
            elif value.__class__.__name__ == cls.__name__:
                return cls(value._payload, value._unpickler, value._object)
            return Mutable.coerce(key, value)
        else:
            return value

    def __repr__(self):
        return repr(self._load())

    def _ipython_key_completions_(self):
        return list(self.keys())


class MappingCell(object):
    __slots__ = ("value", "serialized")

    value: Union[bytes, Any]
    serialized: bool
    _dynamic_loading_threshold = 5e3

    def __init__(self, value, serialized=False):
        self.value = value
        self.serialized = serialized
        if serialized and isinstance(value, bytes) and not starts_with_zstd_magic(value):
            self.value = pyzstd.compress(value)

    def _view_value(self):
        if self.serialized and len(self.value) < self._dynamic_loading_threshold:
            self.deserialize()
        value = self.value if not self.serialized else "<BLOB>"
        return value

    def __repr__(self):
        template = "{self.__class__.__name__}({value}, serialized={self.serialized})"
        value = self._view_value()
        return template.format(self=self, value=value)

    def serialize(self):
        if not self.serialized:
            self.value = pyzstd.compress(dill.dumps(self.value, 2))
            self.serialized = True
        return self

    def deserialize(self):
        if self.serialized:
            if isinstance(self.value, six.text_type):
                self.value = self.value.encode('latin1')
            if starts_with_zstd_magic(self.value):
                self.value = pyzstd.decompress(self.value)
            self.value = dill.loads(self.value)
            self.serialized = False
        return self

    def copy(self):
        return self.__class__(self.value, self.serialized)

    def __reduce__(self):
        if self.serialized:
            return self.__class__, (self.value, self.serialized)
        else:
            return self.copy().serialize().__reduce__()


class PartiallySerializedMutableMapping(MutableMapping[str, MappingCell]):
    def __init__(self, arg=None, **kwargs):
        self.store = {}
        self.update(arg, **kwargs)

    def __getitem__(self, key):
        cell = self.store[key]
        if cell.serialized:
            cell.deserialize()
        return cell.value

    def __setitem__(self, key, value):
        self.store[key] = MappingCell(value, False)

    def __delitem__(self, key):
        del self.store[key]

    def __len__(self):
        return len(self.store)

    def __iter__(self):
        return iter(self.store)

    def __repr__(self):
        rep = {}
        for key, value in self.store.items():
            rep[key] = value._view_value()
        return repr(rep)

    def _ipython_key_completions_(self):
        return list(self.keys())

    def update(self, arg=None, **kwargs):
        if arg:
            if isinstance(arg, PartiallySerializedMutableMapping):
                self.store.update(arg.store)
            else:
                for key, value in arg.items():
                    if isinstance(value, MappingCell):
                        self.store[key] = value
                    else:
                        self.store[key] = MappingCell(value)
        for key, value in kwargs.items():
            if isinstance(value, MappingCell):
                self.store[key] = value
            else:
                self.store[key] = MappingCell(value)

    def unload(self):
        for k, v in self.items():
            if hasattr(v, "serialized"):
                if v.serialized:
                    v.serialize()


class _crossversion_dill(object):

    @staticmethod
    def loads(string, *args, **kwargs):
        try:
            return dill.loads(string, *args, **kwargs)
        except UnicodeDecodeError:
            if six.PY3:
                return dill.loads(string, encoding='latin1')
            else:
                raise

    @staticmethod
    def dumps(obj, *args, **kwargs):
        return dill.dumps(obj, *args, **kwargs)


DillType = partial(PickleType, pickler=_crossversion_dill)


class HasParameters(object):
    def __init__(self, **kwargs):
        self._init_parameters(kwargs)
        super(HasParameters, self).__init__(**kwargs)

    def _init_parameters(self, kwargs):
        kwargs.setdefault('parameters', PartiallySerializedMutableMapping())

    @declared_attr
    def parameters(self) -> LazyMutableMappingWrapper:
        return deferred(Column(LazyMutableMappingWrapper.as_mutable(DillType), default=LazyMutableMappingWrapper))



'''A collection of odds-and-ends that are not heavily used or optimized.
'''

from collections import OrderedDict

import numpy as np
from scipy.ndimage import gaussian_filter1d


class ChromatogramDeltaNode(object):
    '''Represent a sub-region of a chromatogram to determine whether to truncate
    the chromatogram or not based upon whether or not they show large gaps in time
    or significant change in intensity over time.
    '''

    def __init__(self, retention_times, delta_intensity, start_time, end_time, is_below_threshold=True):
        self.retention_times = retention_times
        self.delta_intensity = delta_intensity
        self.start_time = start_time
        self.end_time = end_time
        self.mean_change = np.mean(delta_intensity)
        self.is_below_threshold = is_below_threshold

    def __repr__(self):
        return "ChromatogramDeltaNode(%f, %f, %f)" % (
            self.mean_change, self.start_time, self.end_time)

    @classmethod
    def partition(cls, rt, delta_smoothed, window_size=.5):
        last_rt = rt[1]
        last_index = 1
        nodes = []
        i = 0
        for i, rt_i in enumerate(rt[2:]):
            if (rt_i - last_rt) >= window_size:
                nodes.append(
                    cls(
                        rt[last_index:i],
                        delta_smoothed[last_index:i + 1],
                        last_rt, rt[i]))
                last_index = i
                last_rt = rt_i
        nodes.append(
            cls(
                rt[last_index:i],
                delta_smoothed[last_index:i + 1],
                last_rt, rt[i]))
        return nodes


def build_chromatogram_nodes(rt, signal, sigma=3):
    rt = np.array(rt)
    smoothed = gaussian_filter1d(signal, sigma)
    delta_smoothed = np.gradient(smoothed, rt)
    change = delta_smoothed[:-1] - delta_smoothed[1:]
    avg_change = change.mean()
    std_change = change.std()

    lo = avg_change - std_change
    hi = avg_change + std_change

    nodes = ChromatogramDeltaNode.partition(rt, delta_smoothed)

    for node in nodes:
        if lo > node.mean_change or node.mean_change > hi:
            node.is_below_threshold = False

    return nodes


def find_truncation_points(rt, signal, sigma=3, pad=3):
    nodes = build_chromatogram_nodes(rt, signal, sigma=3)

    leading = 0
    ending = len(nodes)

    for node in nodes:
        if not node.is_below_threshold:
            break
        leading += 1
    leading -= 3
    leading = max(leading - pad, 0)

    for node in reversed(nodes):
        if not node.is_below_threshold:
            break
        ending -= 1

    ending = min(ending + pad, len(nodes) - 1)
    if len(nodes) == 1:
        return nodes[0].start_time, nodes[0].end_time
    elif len(nodes) == 2:
        return nodes[0].start_time, nodes[-1].end_time
    return nodes[leading].start_time, nodes[ending].end_time


class SimpleChromatogram(OrderedDict):
    '''A simplified Chromatogram-like object which supports :meth:`as_arrays`
    and :meth:`get_chromatogram`, but otherwise acts as a mapping from retention
    time to intensity.
    '''
    def __init__(self, *args):
        super(SimpleChromatogram, self).__init__(*args)

    composition = None
    glycan_composition = None

    def as_arrays(self):
        return (
            np.array(list(self.keys())),
            np.array(list(self.values()))
        )

    def get_chromatogram(self):
        return self

    def _new(self):
        return self.__class__()

    def slice(self, start, end):
        pairs = []
        for t, v in self.items():
            if start <= t <= end:
                pairs.append((t, v))
        dup = self._new()
        dup.update(pairs)
        return dup

    def split_sparse(self, delta_rt=1.):
        parts = []
        start = 0
        last = None
        for i, t in enumerate(self.keys()):
            if last is None:
                last = t
                start = t
            if t - last >= delta_rt:
                parts.append(self.slice(start, last))
                start = t
            last = t
        if last != start:
            parts.append(self.slice(start, last))
        return parts

    @property
    def start_time(self):
        return next(iter(self.keys()))

    @property
    def end_time(self):
        return list(self.keys())[-1]

    @property
    def apex_time(self):
        time, intensity = self.as_arrays()
        i = np.argmax(intensity)
        return time[i]


class SimpleEntityChromatogram(SimpleChromatogram):

    def __init__(self, entity=None, glycan_composition=None):
        self.entity = entity
        self.composition = entity
        self.glycan_composition = glycan_composition
        super(SimpleEntityChromatogram, self).__init__()

    def _new(self):
        return self.__class__(self.entity, self.glycan_composition)


class PairedArray(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def as_arrays(self):
        return self.x, self.y

    def __len__(self):
        return len(self.x)



import logging

from typing import Any, Iterable, List, DefaultDict, Optional, Tuple, Union
from glycresoft.serialize.chromatogram import MassShift
from glycresoft.structure.structure_loader import FragmentCachingGlycopeptide


from ms_deisotope.data_source import RandomAccessScanSource, ProcessedScan

from glycresoft import serialize

from glycresoft.chromatogram_tree import (
    Chromatogram, GlycopeptideChromatogram, ChromatogramTreeList,
    ChromatogramFilter)

from glycresoft.trace import ChromatogramExtractor
from glycresoft.scoring import ChromatogramSolution
from glycresoft.tandem.chromatogram_mapping import TandemAnnotatedChromatogram

from glycresoft.tandem.spectrum_match import (
    ScoreSet, FDRSet, MultiScoreSpectrumMatch, MultiScoreSpectrumSolutionSet)

from glycresoft.tandem.glycopeptide.identified_structure import IdentifiedGlycopeptide


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class FakeSpectrumMatch(MultiScoreSpectrumMatch):
    def __init__(self, target, mass_shift=None):
        super(FakeSpectrumMatch, self).__init__(
            scan=None, target=target, score_set=ScoreSet(1e-6, 1e-6, 1e-6, 0),
            best_match=True, data_bundle=None, q_value_set=FDRSet(0.99, 0.99, 0.99, 0.99),
            mass_shift=mass_shift, match_type=0)


class MatchBetweenDataset(object):
    chromatograms: ChromatogramFilter
    identified_structures: List[IdentifiedGlycopeptide]
    label: str

    scan_loader: RandomAccessScanSource[Any, ProcessedScan]
    analysis_loader: Optional[serialize.AnalysisDeserializer]

    _find_by_structure: DefaultDict[Union[str, FragmentCachingGlycopeptide], List[IdentifiedGlycopeptide]]

    def __init__(self, analysis_loader, scan_loader, label=None):
        if label is None and analysis_loader is not None:
            label = analysis_loader.analysis.name

        self.chromatograms = ChromatogramFilter([])
        self.identified_structures = []
        self._find_by_structure = DefaultDict(list)

        self.analysis_loader = analysis_loader
        self.scan_loader = scan_loader
        if self.scan_loader is not None:
            self._load_chromatograms()
        self._prepare_structure_map()
        self.label = label

    @classmethod
    def from_chromatograms_and_identifications(cls, chromatograms: Iterable[Chromatogram],
                                               identifications: List[IdentifiedGlycopeptide],
                                               analysis_loader: Optional[serialize.AnalysisDeserializer]=None,
                                               scan_loader: Optional[RandomAccessScanSource]=None,
                                               label: Optional[str]=None):
        if label is None and analysis_loader is not None:
            label = analysis_loader.analysis.name
        self = cls(None, None, label=label)
        self.analysis_loader = analysis_loader
        self.scan_loader = scan_loader
        self.label = None if analysis_loader is None else analysis_loader.analysis.name

        self.identified_structures = identifications
        self.chromatograms = self._map_chromatograms(chromatograms, identifications)

        self._prepare_structure_map()
        return self

    def _map_chromatograms(self, chromatograms: Iterable[Chromatogram],
                           identifications: Iterable[IdentifiedGlycopeptide]):
        for chrom in chromatograms:
            chrom.mark = False
        for idgp in identifications:
            if idgp.chromatogram is None:
                continue
            for mshift in idgp.mass_shifts:
                chroma = chromatograms.find_all_by_mass(
                    idgp.weighted_neutral_mass + mshift.mass, 1e-5)
                for chrom in chroma:
                    if idgp.chromatogram.overlaps_in_time(chrom):
                        chrom.mark = True

        chromatograms = ChromatogramFilter(
            [chrom for chrom in chromatograms if not chrom.mark] + list(identifications))
        return chromatograms

    def _load_chromatograms(self):
        logger.info(f"Loading chromatograms from {self.scan_loader.source_file_name}")
        extractor = ChromatogramExtractor(
            self.scan_loader, minimum_mass=1000.0, grouping_tolerance=1.5e-5)
        chromatograms = extractor.run()
        for chrom in chromatograms:
            chrom.mark = False
        if self.analysis_loader is not None:
            logger.info(f"Loading identified glycopeptides from {self.analysis_loader.analysis.name}")
            idgps = self.analysis_loader.load_identified_glycopeptides(0.1)
            self.identified_structures = idgps
            self.chromatograms = self._map_chromatograms(chromatograms, idgps)

    @property
    def ms1_scoring_model(self):
        return self.analysis_loader.analysis.parameters.get('scoring_model')

    def find(self, ids: IdentifiedGlycopeptide,
             mass_error_tolerance: float=1e-5,
             time_error_tolerance: float=2.0) -> List[Tuple[Union[Chromatogram, IdentifiedGlycopeptide],
                                                      Optional[MassShift]]]:
        key = ids.structure
        out = []
        id_out = []
        ids_mass = ids.weighted_neutral_mass
        ids_apex_time = ids.apex_time
        if ids_apex_time is None:
            ids_apex_time = ids.tandem_solutions[0].scan_time
        if key in self._find_by_structure:
            results = self._find_by_structure[key]
            for result in results:
                if abs(result.weighted_neutral_mass - ids_mass) / ids_mass < mass_error_tolerance:
                    apex_time = result.apex_time
                    if apex_time is None:
                        apex_time = result.tandem_solutions[0].scan_time
                    if abs(apex_time - ids_apex_time) < time_error_tolerance:
                        id_out.append((result, None))
        for mshift in ids.mass_shifts:
            qmass = mshift.mass + ids_mass
            chroma = self.chromatograms.find_all_by_mass(
                qmass, mass_error_tolerance)
            for chrom in chroma:
                if abs(chrom.apex_time - ids_apex_time) < time_error_tolerance:
                    if not isinstance(chrom, Chromatogram) and (chrom, None) in id_out:
                        continue
                    out.append((chrom, mshift))
        return id_out + out

    def get_identified_structure_for(self, structure):
        key = structure
        candidates = self._find_by_structure[key]
        for candidate in candidates:
            if candidate.structure == structure:
                return candidate
        raise KeyError("Could not locate %r (%r)" %
                        (structure, structure.protein_relation))

    def create(self, structure, chromatogram, shift):
        chrom = GlycopeptideChromatogram(
            structure, ChromatogramTreeList())
        chrom = chrom.merge(chromatogram, shift)

        # If there is another identification that this wasn't merged with because of varying
        # errors in apex time matching, things break down. Check just in case we really want
        # to merge here.
        try:
            existing = self.get_identified_structure_for(structure)
            if existing.chromatogram.common_nodes(chrom):
                return None
            else:
                return self.merge(structure, chromatogram, shift)
        except KeyError:
            pass
        ms1_model = self.ms1_scoring_model
        chrom = ChromatogramSolution(
            chrom, ms1_model.logitscore(chrom), scorer=ms1_model)
        chrom = TandemAnnotatedChromatogram(chrom)
        sset = MultiScoreSpectrumSolutionSet(
            None, [FakeSpectrumMatch(structure)])
        sset.q_value = sset.best_solution().q_value
        idgp = IdentifiedGlycopeptide(structure, [sset], chrom)
        return idgp

    def merge(self, structure, chromatogram, shift):
        candidate = self.get_identified_structure_for(structure)
        chrom = GlycopeptideChromatogram(
            structure, ChromatogramTreeList())
        chrom = chrom.merge(chromatogram, shift)
        chrom = TandemAnnotatedChromatogram(chrom)
        if candidate.chromatogram is None:
            candidate.chromatogram = chrom
        else:
            candidate.chromatogram = candidate.chromatogram.merge(chrom)

    def add(self, ids):
        key = ids.structure
        self._find_by_structure[key].append(ids)
        self.identified_structures.append(ids)
        self.chromatograms.extend([ids])

    def _protein_name_label_map(self):
        proteins = self.analysis_loader.query(serialize.Protein).all()
        name_map = dict()
        for prot in proteins:
            name_map[prot.id] = prot.name
        return name_map

    def _prepare_structure_map(self):
        self._find_by_structure = DefaultDict(list)
        if self.analysis_loader is not None:
            name_map = self._protein_name_label_map()
        else:
            name_map = None
        for ids in self.identified_structures:
            try:
                if name_map is not None:
                    pr = ids.protein_relation
                    pr.protein_id = name_map.get(pr.protein_id, pr.protein_id)
            except AttributeError:
                pass
            self._find_by_structure[(ids.structure)].append(ids)



from collections import Counter

from sqlalchemy import (
    Column, Integer, String, ForeignKey, PickleType, func)

from sqlalchemy.orm import relationship, object_session
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import MutableDict

from glycresoft.serialize.base import (
    Base, HasUniqueName)

from .generic import HasFiles


from glypy.structure.glycan_composition import FrozenGlycanComposition


class HypothesisBase(HasUniqueName, HasFiles):
    @declared_attr
    def id(self):
        return Column(Integer, primary_key=True, autoincrement=True)

    parameters = Column(MutableDict.as_mutable(PickleType), default=dict(), nullable=False)
    status = Column(String(28))

    def __repr__(self):
        return "{self.__class__.__name__}(id={self.id}, name={self.name})".format(self=self)


class GlycanHypothesis(HypothesisBase, Base):
    __tablename__ = "GlycanHypothesis"

    def monosaccharide_bounds(self):
        from .glycan import GlycanComposition
        bounds = Counter()
        session = object_session(self)
        for composition, in session.query(GlycanComposition.composition.distinct()).filter(
                GlycanComposition.hypothesis_id == self.id):
            composition = FrozenGlycanComposition.parse(composition)
            for residue, count in composition.items():
                bounds[residue] = max(bounds[residue], count)
        return {str(k): int(v) for k, v in bounds.items()}

    @property
    def glycan_class_counts(self):
        from .glycan import GlycanClass, GlycanComposition
        session = object_session(self)
        d = {k: v for k, v in session.query(GlycanClass.name, func.count(GlycanClass.name)).join(
             GlycanComposition.structure_classes).group_by(GlycanClass.name).all()}
        return d

    @property
    def n_glycan_only(self):
        counts = self.glycan_class_counts
        k = counts.get('N-Glycan', 0)
        return len(counts) == 1 and k > 0


class GlycopeptideHypothesis(HypothesisBase, Base):
    __tablename__ = "GlycopeptideHypothesis"

    glycan_hypothesis_id = Column(Integer, ForeignKey(GlycanHypothesis.id, ondelete="CASCADE"), index=True)
    glycan_hypothesis = relationship(GlycanHypothesis)

    def monosaccharide_bounds(self):
        return self.glycan_hypothesis.monosaccharide_bounds()

    @property
    def glycan_class_counts(self):
        return self.glycan_hypothesis.glycan_class_counts

    @property
    def n_glycan_only(self):
        return self.glycan_hypothesis.n_glycan_only


from itertools import cycle
from matplotlib.colors import cnames, hex2color, rgb_to_hsv, hsv_to_rgb
from matplotlib import patches as mpatches


def lighten(rgb, factor=0.25):
    '''Given a triplet of rgb values, lighten the color by `factor`%'''
    factor += 1
    return [min(c * factor, 1) for c in rgb]


def darken(rgb, factor=0.25):
    '''Given a triplet of rgb values, darken the color by `factor`%'''
    factor = 1 - factor
    return [(c * factor) for c in rgb]


def deepen(rgb, factor=0.25):
    hsv = rgb_to_hsv(rgb)
    hsv[1] = min(hsv[1] * (1 + factor), 1)
    return hsv_to_rgb(hsv)


colors = cycle([hex2color(cnames[name]) for name in (
    "red", "blue", "yellow", "purple", "navy", "grey", "coral", "forestgreen", "limegreen", "maroon", "aqua",
    "lavender", "lightcoral", "mediumorchid")])


class ColorMapper(object):
    colors = [hex2color(cnames[name]) for name in (
              "red", "blue", "yellow", "purple", "navy", "grey", "coral", "forestgreen", "limegreen", "maroon", "aqua",
              "lavender", "lightcoral", "mediumorchid")]

    def __init__(self):
        self.color_name_map = {
            "HexNAc": hex2color(cnames["mediumseagreen"]),
            "N-Glycosylation": hex2color(cnames["mediumseagreen"]),
            "O-Glycosylation": hex2color(cnames["cadetblue"]),
            "GAG-Linker": hex2color(cnames['burlywood']),
            "Xyl": hex2color(cnames['darkorange']),
            "a,enHex": hex2color(cnames['lightskyblue']),
            "aHex": hex2color(cnames['plum']),
            "Hex": hex2color(cnames['steelblue']),
            "Neu5Ac": hex2color(cnames['blueviolet']),
            "Neu5Gc": hex2color(cnames['lightsteelblue']),
            "Fuc": hex2color(cnames['crimson'])
        }
        self.color_generator = cycle(self.colors)

    def get_color(self, name):
        """Given a name, find the color mapped to that name, or
        select the next color from the `colors` generator and assign
        it to the name and return the new color.

        Parameters
        ----------
        name : object
            Any hashable object, usually a string

        Returns
        -------
        tuple: RGB triplet
        """
        try:
            return self.color_name_map[name]
        except KeyError:
            o = self.color_name_map[name] = next(self.color_generator)
            return o

    __getitem__ = get_color

    def __setitem__(self, name, color):
        self.color_name_map[name] = color

    def __repr__(self):
        return repr(self.color_name_map)

    darken = staticmethod(darken)
    lighten = staticmethod(lighten)

    def keys(self):
        return self.color_name_map.keys()

    def items(self):
        return self.color_name_map.items()

    def proxy_artists(self, subset=None):
        proxy_artists = []
        if subset is not None:
            for name, color in self.items():
                if name in subset:
                    artist = mpatches.Rectangle((0, 0), 1, 1, fc=color, label=name)
                    proxy_artists.append((name, artist))
        else:
            for name, color in self.items():
                artist = mpatches.Rectangle((0, 0), 1, 1, fc=color, label=name)
                proxy_artists.append((name, artist))
        return proxy_artists


_color_mapper = ColorMapper()

color_name_map = _color_mapper.color_name_map
get_color = _color_mapper.get_color
proxy_artists = _color_mapper.proxy_artists


def color_dict():
    return {str(k): v for k, v in _color_mapper.items()}


material_palette = [(0.9568627450980393, 0.2627450980392157, 0.21176470588235294),
                    (0.24705882352941178, 0.3176470588235294, 0.7098039215686275),
                    (0.2980392156862745, 0.6862745098039216, 0.3137254901960784),
                    (1.0, 0.596078431372549, 0.0),
                    (0.9137254901960784, 0.11764705882352941, 0.38823529411764707),
                    (0.12941176470588237, 0.5882352941176471, 0.9529411764705882),
                    (0.5450980392156862, 0.7647058823529411, 0.2901960784313726),
                    (1.0, 0.3411764705882353, 0.13333333333333333),
                    (0.011764705882352941, 0.6627450980392157, 0.9568627450980393),
                    (0.803921568627451, 0.8627450980392157, 0.2235294117647059),
                    (0.4745098039215686, 0.3333333333333333, 0.2823529411764706),
                    (0.611764705882353, 0.15294117647058825, 0.6901960784313725),
                    (0.0, 0.7372549019607844, 0.8313725490196079),
                    (1.0, 0.9215686274509803, 0.23137254901960785),
                    (0.6196078431372549, 0.6196078431372549, 0.6196078431372549),
                    (0.403921568627451, 0.22745098039215686, 0.7176470588235294),
                    (0.0, 0.5882352941176471, 0.5333333333333333),
                    (1.0, 0.7568627450980392, 0.027450980392156862),
                    (0.3764705882352941, 0.49019607843137253, 0.5450980392156862)]



'''
Usage:
    list_of_db_glycans = get_db_glycans()
    gcf = GlycanCompositionFilter(list_of_db_glycans)
    index = InclusionFilter(gcf.query("Fuc", 1, 3).add("HexNAc", 4, 6).add("Neu5Ac", 0, 2))
    id_of_interest = 5
    print(5 in index)
'''
from collections import defaultdict, Counter


class FilterTreeNode(object):
    def __init__(self, members=None):
        if members is None:
            members = []
        self.members = members
        self.children = defaultdict(FilterTreeNode)
        self.splitting_key = None
        self.split_value = None

    def add(self, member):
        self.members.append(member)

    def split_on(self, key):
        if self.splitting_key is not None:
            raise ValueError("Already split node")
        for member in self.members:
            self.children[member[key]].add(member)

        self.splitting_key = key
        for key, child in self.children.items():
            child.split_value = key

    def split_sequence(self, key_sequence):
        if len(key_sequence) == 0:
            return
        key = key_sequence[0]
        self.split_on(key)
        for child in self.children.values():
            child.split_sequence(key_sequence[1:])

    def __repr__(self):
        return "FilterTreeNode(%d, %r, ^%r, %d)" % (
            len(self.members), self.splitting_key,
            self.split_value, len(self.children))

    def query(self, key, lo=0, hi=100):
        if str(key) == str(self.splitting_key):
            result_set = []
            hi = min(max(self.children.keys()), hi)
            for i in range(lo, hi + 1):
                result_set.append(self.children[i])
            return QuerySet(result_set)
        else:
            out = []
            for child in self.children.values():
                out.append(child.query(key, lo, hi))
            return QuerySet.union(out)

    def __iter__(self):
        if len(self.children) == 0:
            for member in self.members:
                yield member
        else:
            for child in self.children.values():
                for item in child:
                    yield item


class QuerySet(object):
    def __init__(self, node_list):
        self.node_list = node_list

    def __iter__(self):
        for node in self.node_list:
            for item in node:
                yield item

    def query(self, key, lo=0, hi=100):
        out = []
        for node in self.node_list:
            out.append(node.query(key, lo, hi))
        return QuerySet.union(out)

    @classmethod
    def union(cls, queries):
        node_list = []
        for q in queries:
            node_list.extend(q.node_list)
        return cls(node_list)

    @classmethod
    def intersect(cls, queries):
        item_map = {}
        member_map = Counter()
        for q in queries:
            for item in q:
                item_map[item.id] = item
                member_map[item.id] += 1
        n = len(queries)
        out = []
        for key, value in member_map.items():
            if value == n:
                out.append(item_map[key])
        return out


class QueryInterval(object):
    def __init__(self, key, low=0, high=100):
        self.key = str(key)
        self.low = low
        self.high = high

    def __repr__(self):
        return "QueryInterval(%r, %d, %d)" % (self.key, self.low, self.high)

    def __iter__(self):
        yield self.key
        yield self.low
        yield self.high

    def __eq__(self, other):
        return (self.key == other.key and self.low == other.low and self.high == other.high)

    def __ne__(self, other):
        return not (self == other)


class QueryComposer(object):
    def __init__(self, parent):
        self.parent = parent
        self.filters = []

    def __call__(self, filter_spec_or_key, lo=0, hi=100):
        if isinstance(filter_spec_or_key, QueryInterval):
            self.filters.append(filter_spec_or_key)
        else:
            self.filters.append(QueryInterval(filter_spec_or_key, lo, hi))
        return self

    def add(self, *args, **kwargs):
        self(*args, **kwargs)
        return self

    def collate(self):
        monosaccharides = list(map(str, self.parent.monosaccharides))
        filters = sorted(self.filters, key=lambda x: str(x.key))
        kept_filters = []
        for filt in filters:
            if filt.key in monosaccharides:
                kept_filters.append(filt)
        return kept_filters

    def __eq__(self, other):
        if isinstance(other, QueryComposer):
            return self.collate() == other.collate()
        else:
            return self.collate() == other

    def __ne__(self, other):
        return not (self == other)

    def all(self):
        return list(self._compose())

    def _compose(self):
        query = None
        for filt in self.collate():
            if query is None:
                query = self.parent.root.query(*filt)
            else:
                query = query.query(*filt)
        if query is None:
            return self.parent
        return query

    def __iter__(self):
        return iter(self._compose())


class GlycanCompositionFilter(object):
    def __init__(self, members):
        self.members = members
        self.monosaccharides = list()
        self._extract_monosaccharides(members)
        self.root = FilterTreeNode(members)
        self._build_partitions()

    def _extract_monosaccharides(self, iterable):
        monosaccharides = set()
        for case in iterable:
            monosaccharides.update(case.keys())
        self.monosaccharides = sorted(monosaccharides, key=str)

    def _build_partitions(self):
        self.root.split_sequence(self.monosaccharides)

    def query(self, *args, **kwargs):
        compose = QueryComposer(self)
        return compose(*args, **kwargs)

    def filter_index(self, *args, **kwargs):
        query_set = self.query(*args, **kwargs)
        return InclusionFilter(query_set)

    def __iter__(self):
        return iter(self.members)


class InclusionFilter(object):
    def __init__(self, query_set):
        self.query_set = query_set
        self.hash_index = set()
        self._build_index()

    def _build_index(self):
        for case in self.query_set:
            self.hash_index.add(case.id)

    def __repr__(self):
        return "InclusionFilter()"

    def contains_all(self, keys):
        for key in keys:
            if key not in self:
                return False
        return True

    def __contains__(self, key):
        return key in self.hash_index



from sqlalchemy import exc, func
from sqlalchemy.orm.session import object_session

from glycresoft.serialize.base import Base

from glycresoft.serialize.connection import DatabaseBoundOperation, ConnectFrom

from glycresoft.serialize.spectrum import (
    SampleRun,
    MSScan,
    PrecursorInformation,
    DeconvolutedPeak)

from glycresoft.serialize.analysis import (
    Analysis,
    BoundToAnalysis,
    AnalysisTypeEnum)

from glycresoft.serialize.chromatogram import (
    ChromatogramTreeNode,
    Chromatogram,
    ChromatogramSolution,
    MassShift,
    CompositionGroup,
    CompoundMassShift,
    UnidentifiedChromatogram,
    GlycanCompositionChromatogram,
    ChromatogramSolutionMassShiftedToChromatogramSolution)

from glycresoft.serialize.tandem import (
    GlycopeptideSpectrumCluster,
    GlycopeptideSpectrumMatch,
    GlycopeptideSpectrumSolutionSet,
    GlycanCompositionSpectrumCluster,
    GlycanCompositionSpectrumSolutionSet,
    GlycanCompositionSpectrumMatch,
    UnidentifiedSpectrumCluster,
    UnidentifiedSpectrumSolutionSet,
    UnidentifiedSpectrumMatch,
    GlycopeptideSpectrumMatchScoreSet,
    )

from glycresoft.serialize.identification import (
    IdentifiedGlycopeptide,
    AmbiguousGlycopeptideGroup)

from glycresoft.serialize.serializer import (
    AnalysisSerializer,
    AnalysisDeserializer,
    DatabaseScanDeserializer,
    DatabaseBoundOperation)

from glycresoft.serialize.hypothesis import *

# from glycresoft.serialize.migration import (
#     GlycanCompositionChromatogramAnalysisSerializer,
#     GlycopeptideMSMSAnalysisSerializer)

from glycresoft.serialize import config

import sys
from glycresoft.serialize import spectrum

sys.modules['ms_deisotope.output.db'] = spectrum



from typing import Any, Callable, Dict, List, Tuple, Type, NamedTuple

from ms_deisotope.data_source import ProcessedRandomAccessScanSource, ProcessedScan

from glycresoft.task import TaskBase

from glycresoft.chromatogram_tree import Chromatogram, ChromatogramFilter, MassShift, mass_shift


from glycresoft.tandem.target_decoy import FDREstimatorBase

from glycresoft.tandem.glycopeptide.identified_structure import IdentifiedGlycopeptide

from glycresoft.tandem.spectrum_match.spectrum_match import MultiScoreSpectrumMatch, SpectrumMatch, SpectrumMatcherBase
from glycresoft.tandem.spectrum_match import SpectrumSolutionSet, MultiScoreSpectrumSolutionSet


class CoElutionCandidate(NamedTuple):
    reference: IdentifiedGlycopeptide
    extensions: List[Chromatogram]
    ms2_spectra: List[MultiScoreSpectrumMatch]
    mass_shift: MassShift

    def merge(self, scan_id_to_solution_set: Dict[str, SpectrumSolutionSet]) -> Tuple[IdentifiedGlycopeptide, List[IdentifiedGlycopeptide]]:
        ssets = {}
        to_drop = []
        existing_matches = {sset.scan.scan_id for sset in self.reference.spectrum_matches}
        for gpsm in self.ms2_spectra:
            sset = None
            if gpsm.scan_id in scan_id_to_solution_set:
                sset = scan_id_to_solution_set[gpsm.scan_id]
                sset.append(gpsm)
            else:
                sset = MultiScoreSpectrumSolutionSet(gpsm.scan, [gpsm])
                scan_id_to_solution_set[gpsm.scan_id] = sset
            if gpsm.scan_id not in existing_matches:
                ssets[gpsm.scan.scan_id] = sset

        for part in self.extensions:
            if not isinstance(part, IdentifiedGlycopeptide):
                self.reference.chromatogram.merge_in_place(part, node_type=self.mass_shift)
            else:
                to_drop.append(part)
                self.reference.chromatogram.merge_in_place(
                    part.chromatogram, node_type=self.mass_shift)
        self.reference.tandem_solutions.extend(ssets.values())
        return self.reference, to_drop


class CoElutionAdductFinder(TaskBase):
    scan_loader: ProcessedRandomAccessScanSource
    chromatograms: ChromatogramFilter

    msn_scoring_model: Type[SpectrumMatcherBase]
    msn_evaluation_args: Dict[str, Any]
    fdr_estimator: FDREstimatorBase

    scan_id_to_solution_set: Dict[str, SpectrumSolutionSet]
    threshold_fn: Callable[[SpectrumMatch], bool]

    def __init__(self, scan_loader: ProcessedRandomAccessScanSource,
                 chromatograms: ChromatogramFilter,
                 msn_scoring_model: SpectrumMatcherBase,
                 msn_evaluation_args: Dict[str, Any],
                 fdr_estimator: FDREstimatorBase,
                 scan_id_to_solution_set: Dict[str, SpectrumSolutionSet],
                 threshold_fn: Callable[[SpectrumMatch], bool]=lambda x: x.q_value < 0.05):
        self.scan_loader = scan_loader
        self.chromatograms = chromatograms

        self.msn_scoring_model = msn_scoring_model
        self.msn_evaluation_args = msn_evaluation_args

        self.fdr_estimator = fdr_estimator
        self.threshold_fn = threshold_fn

        self.scan_id_to_solution_set = scan_id_to_solution_set

    def find_coeluting(
        self,
        reference: IdentifiedGlycopeptide,
        mass: float,
        error_tolerance: float = 2e-5,
    ) -> List[Chromatogram]:
        alternates = [
            alt for alt in self.chromatograms.find_all_by_mass(mass, error_tolerance)
            if reference.overlaps_in_time(alt)
        ]
        return alternates

    def find_ms2_spectra_for_chromatogram(self, chromatogram: Chromatogram,
                                          error_tolerance: float) -> List[ProcessedScan]:
        pinfos = self.scan_loader.msms_for(
            chromatogram.weighted_neutral_mass,
            error_tolerance,
            start_time=chromatogram.start_time - 3,
            end_time=chromatogram.end_time + 3,
        )
        return [p.product for p in pinfos]

    def evaluate_msn_spectra(self, reference: IdentifiedGlycopeptide,
                             adduct: MassShift,
                             product_spectra: List[ProcessedScan]) -> List[MultiScoreSpectrumMatch]:
        acc = []
        expects_tandem_shift = adduct.tandem_mass != 0
        for product in product_spectra:
            match: SpectrumMatcherBase = self.msn_scoring_model.evaluate(
                product,
                reference.structure,
                mass_shift=adduct,
                **self.msn_evaluation_args
            )

            modified_peaks = []
            for pfp in match.solution_map:
                chem_shift = pfp.fragment.chemical_shift
                if chem_shift and chem_shift.name == adduct.name:
                    modified_peaks.append(pfp)

            if expects_tandem_shift and not modified_peaks:
                continue

            match: MultiScoreSpectrumMatch = MultiScoreSpectrumMatch.from_match_solution(match)

            self.fdr_estimator.score_all([match])
            if product.scan_id in self.scan_id_to_solution_set:
                sset = self.scan_id_to_solution_set[product.scan_id]
                alt_match = sset.best_solution()
                if alt_match.score > match.score:
                    continue

            if self.threshold_fn(match):
                acc.append(match)
        return acc

    def process_candidates(self, identified_structures: List[IdentifiedGlycopeptide],
                           candidate_updates: List[CoElutionCandidate]) -> List[IdentifiedGlycopeptide]:
        to_remove = set()
        updated_identifications = []

        for update_candidate in candidate_updates:
            updated, dropped = update_candidate.merge(self.scan_id_to_solution_set)
            to_remove.update(dropped)
            updated_identifications.append(updated)

        for idgp_to_remove in to_remove:
            identified_structures.remove(idgp_to_remove)
        return identified_structures

    def coeluting_adduction_candidates(
        self,
        identified_structures: List[IdentifiedGlycopeptide],
        adduct: MassShift,
        error_tolerance: float = 1e-5,
    ) -> List[CoElutionCandidate]:
        results = []
        n = len(identified_structures)
        for i, ref in enumerate(identified_structures):
            if i % 250 == 0:
                self.log(f"{i}/{n} ({i / n * 100.0:0.2f}%) {len(results)} cases found")
            if ref.chromatogram is None:
                continue
            alts = self.find_coeluting(
                ref,
                ref.weighted_neutral_mass + adduct.mass,
                error_tolerance
            )
            tandem_candidates: List[ProcessedScan] = []
            for alt in alts:
                tandem_candidates.extend(
                    self.find_ms2_spectra_for_chromatogram(alt, error_tolerance))
            acc = self.evaluate_msn_spectra(ref, adduct, tandem_candidates)
            if acc:
                results.append(
                    CoElutionCandidate(ref, alts, acc, adduct))
        return results

    def handle_adduct(self, identified_structures: List[IdentifiedGlycopeptide],
                      adduct: MassShift,
                      error_tolerance: float = 1e-5) -> List[IdentifiedGlycopeptide]:
        candidates = self.coeluting_adduction_candidates(
            identified_structures, adduct, error_tolerance)
        identified_structures = self.process_candidates(
            identified_structures, candidates)
        return identified_structures



from .analysis_migration import (
    GlycanCompositionChromatogramAnalysisSerializer,
    GlycopeptideMSMSAnalysisSerializer,
    DynamicGlycopeptideMSMSAnalysisSerializer)


__all__ = [
    "GlycopeptideMSMSAnalysisSerializer",
    "DynamicGlycopeptideMSMSAnalysisSerializer",
    "GlycanCompositionChromatogramAnalysisSerializer"
]



from collections import defaultdict, Counter
import itertools

from six import string_types as basestring

import glypy
from glypy.structure.glycan_composition import FrozenGlycanComposition
from glypy.composition import formula
from glypy import ReducedEnd

from .glycan_source import GlycanHypothesisSerializerBase
from glycresoft.serialize import (
    DatabaseBoundOperation,
    GlycanComposition as DBGlycanComposition,
    GlycanCompositionToClass)


def get_glycan_composition(glycan_composition):
    try:
        return glycan_composition.convert()
    except AttributeError:
        return glycan_composition


def linearize_composition(glycan_composition):
    return [mr for mr, count in glycan_composition.items() for i in range(count)]


INF = float('inf')


class GlycanCompositionTransformerRule(object):
    def __init__(self, modifier, constraints):
        self.modifier = modifier
        self.constraints = constraints

    def __repr__(self):
        return "%s(%r, %r)" % (
            self.__class__.__name__,
            self.modifier, self.constraints)

    def test(self, glycan_composition):
        return self.constraints(glycan_composition)

    def apply(self, glycan_composition, i):
        glycan_composition[self.modifier] = i
        return glycan_composition

    def __eq__(self, other):
        return self.modifier == other.modifier and self.constraints == other.constraints

    def __hash__(self):
        return hash(self.modifier)


class SiteSpecificGlycanCompositionTransformerRule(object):
    def __init__(self, target, modifier):
        self.target = target
        self.modifier = modifier

    def find_valid_sites(self, linear):
        return [i for i, mr in enumerate(linear) if mr == self.target]

    def __repr__(self):
        return "%s(%r, %r)" % (
            self.__class__.__name__,
            self.target, self.modifier)

    def apply(self, glycan_composition, i):
        glycan_composition[self.modifier] = i
        return glycan_composition

    def __eq__(self, other):
        return self.target == other.target and self.modifier == other.modifier

    def __hash__(self):
        return hash(self.target)


class GlycanReductionTransformerRule(object):
    def __init__(self, reduction):
        if isinstance(reduction, ReducedEnd):
            reduction = reduction.clone()
        elif isinstance(reduction, glypy.Composition):
            reduction = ReducedEnd(reduction.clone())
        elif isinstance(reduction, basestring):
            reduction = ReducedEnd(glypy.Composition(reduction))
        self.reduction = reduction

    def apply(self, glycan_composition):
        glycan_composition.reducing_end = self.reduction.clone()
        return glycan_composition


def split_reduction_and_modification_rules(rule_list):
    modification_rules = []
    reduction_rules = []
    for rule in rule_list:
        if isinstance(rule, GlycanReductionTransformerRule):
            reduction_rules.append(rule)
        else:
            modification_rules.append(rule)
    return modification_rules, reduction_rules


def modification_series(variable_sites):
    """Given a dictionary mapping between modification names and
    an iterable of valid sites, create a dictionary mapping between
    modification names and a list of valid sites plus the constant `None`

    Parameters
    ----------
    variable_sites : dict
        Description

    Returns
    -------
    dict
        Description
    """
    sites = defaultdict(list)
    for mod, varsites in variable_sites.items():
        for site in varsites:
            sites[site].append(mod)
    for site in list(sites):
        sites[site].append(None)
    return sites


def site_modification_assigner(modification_sites_dict):
    sites = modification_sites_dict.keys()
    choices = modification_sites_dict.values()
    for selected in itertools.product(*choices):
        yield zip(sites, selected)


def simplify_assignments(assignments_generator):
    distinct = set()
    for assignments in assignments_generator:
        distinct.add(frozenset(Counter(mod for site, mod in assignments if mod is not None).items()))
    return distinct


def glycan_composition_permutations(glycan_composition, constant_modifications=None, variable_modifications=None):
    if constant_modifications is None:
        constant_modifications = []
    if variable_modifications is None:
        variable_modifications = []

    glycan_composition = FrozenGlycanComposition(
        get_glycan_composition(glycan_composition))
    final_composition = FrozenGlycanComposition()
    temporary_composition = FrozenGlycanComposition()
    sequence = linearize_composition(glycan_composition)

    if not constant_modifications:
        temporary_composition = glycan_composition.clone()
        has_constant_reduction = False
    else:
        constant_modifications, constant_reduction = split_reduction_and_modification_rules(
            constant_modifications)
        for rule in constant_modifications:
            extracted = rule.find_valid_sites(sequence)
            for i, mr in enumerate(sequence):
                if i in extracted:
                    final_composition[mr] += 1
                    final_composition[rule.modifier] += 1
                else:
                    temporary_composition[mr] += 1
            sequence = linearize_composition(temporary_composition)
        if constant_reduction:
            has_constant_reduction = True
            constant_reduction[0].apply(temporary_composition)

    variable_modifications, variable_reduction = split_reduction_and_modification_rules(
        variable_modifications)

    mod_site_map = {
        rule: rule.find_valid_sites(sequence) for rule in variable_modifications
    }
    assignments_generator = site_modification_assigner(
        modification_series(mod_site_map))
    for assignments in simplify_assignments(assignments_generator):
        result = temporary_composition.clone()
        for mod, count in assignments:
            mod.apply(result, count)
        result += final_composition
        if glycan_composition.reducing_end is not None:
            result.reducing_end = glycan_composition.reducing_end.clone()
        yield result
        if not has_constant_reduction:
            for rule in variable_reduction:
                yield rule.apply(result.clone())


class GlycanCompositionPermutationHypothesisSerializer(GlycanHypothesisSerializerBase):
    def __init__(self, source_hypothesis_id, database_connection, constant_modifications=None,
                 variable_modifications=None, hypothesis_name=None):
        super(GlycanCompositionPermutationHypothesisSerializer, self).__init__(database_connection, hypothesis_name)
        self.source_hypothesis_id = source_hypothesis_id
        self.constant_modifications = constant_modifications
        self.variable_modifications = variable_modifications

        self.loader = None
        self.transformer = None

    def make_pipeline(self):
        self.loader = iter(DatabaseBoundOperation(
            self._original_connection).query(
            DBGlycanComposition).filter(
            DBGlycanComposition.hypothesis_id == self.source_hypothesis_id).all())
        self.transformer = self.permuter(self.loader)

    def permuter(self, source_iterator):
        for glycan_composition in source_iterator:
            structure_classes = [sc.name for sc in glycan_composition.structure_classes]
            for permutation in glycan_composition_permutations(
                    glycan_composition, self.constant_modifications, self.variable_modifications):
                yield permutation, structure_classes

    def run(self):
        self.make_pipeline()
        structure_class_lookup = self.structure_class_loader
        self.log("Loading Glycan Compositions from Stream for %r" % self.hypothesis)

        acc = []
        for composition, structure_classes in self.transformer:
            mass = composition.mass()
            composition_string = composition.serialize()
            formula_string = formula(composition.total_composition())
            inst = DBGlycanComposition(
                calculated_mass=mass, formula=formula_string,
                composition=composition_string,
                hypothesis_id=self.hypothesis_id)
            self.session.add(inst)
            self.session.flush()
            for structure_class in structure_classes:
                structure_class = structure_class_lookup[structure_class]
                acc.append(dict(glycan_id=inst.id, class_id=structure_class.id))
                if len(acc) % 100 == 0:
                    self.session.execute(GlycanCompositionToClass.insert(), acc)
                    acc = []
        if acc:
            self.session.execute(GlycanCompositionToClass.insert(), acc)
            acc = []
        self.session.commit()



import os

from collections import defaultdict

import dill as pickle

import numpy as np

from scipy.stats import gaussian_kde

import glycopeptidepy

from glycresoft.task import TaskBase
from glycresoft.output.csv_format import csv_stream

from .structure import GlycopeptideChromatogramProxy
from .cross_run import ReplicatedAbundanceWeightedPeptideFactorElutionTimeFitter


# Deprecated. Adapt to model.GlycopeptideElutionTimeModelBuildingPipeline
class GlycopeptideElutionTimeModeler(TaskBase):
    _model_class = ReplicatedAbundanceWeightedPeptideFactorElutionTimeFitter

    def __init__(self, glycopeptide_chromatograms, factors=None, refit_filter=0.01, replicate_key_attr=None,
                 test_chromatograms=None, use_retention_time_normalization=False, prefer_joint_model=False,
                 minimum_observations_for_specific_model=20):
        if replicate_key_attr is None:
            replicate_key_attr = 'analysis_name'
        if test_chromatograms is not None:
            if not isinstance(test_chromatograms[0], GlycopeptideChromatogramProxy):
                test_chromatograms = [
                    GlycopeptideChromatogramProxy.from_chromatogram(i) for i in test_chromatograms]

        self.replicate_key_attr = replicate_key_attr
        if not isinstance(glycopeptide_chromatograms[0], GlycopeptideChromatogramProxy):
            glycopeptide_chromatograms = [
                GlycopeptideChromatogramProxy.from_chromatogram(i) for i in glycopeptide_chromatograms]
        self.glycopeptide_chromatograms = glycopeptide_chromatograms
        self.test_chromatograms = test_chromatograms
        self.prefer_joint_model = prefer_joint_model
        self.minimum_observations_for_specific_model = minimum_observations_for_specific_model
        self.factors = factors
        if self.factors is None:
            self.factors = self._infer_factors()
        self.use_retention_time_normalization = use_retention_time_normalization
        self.joint_model = None
        self.refit_filter = refit_filter
        self.by_peptide = defaultdict(list)
        self.peptide_specific_models = dict()
        self.delta_by_factor = dict()
        self._partition_by_sequence()

    def _partition_by_sequence(self):
        for record in self.glycopeptide_chromatograms:
            key = glycopeptidepy.parse(str(record.structure)).deglycosylate()
            self.by_peptide[key].append(record)

    def _deltas_for(self, monosaccharide):
        deltas = []
        for _backbone, cases in self.by_peptide.items():
            for target in cases:
                gc = target.glycan_composition.clone()
                gc[monosaccharide] += 1
                key = self.joint_model._get_replicate_key(target)
                for case in cases:
                    if case.glycan_composition == gc and self.joint_model._get_replicate_key(case) == key:
                        deltas.append(case.apex_time - target.apex_time)
        return np.array(deltas)

    def _infer_factors(self):
        keys = set()
        for record in self.glycopeptide_chromatograms:
            keys.update(record.glycan_composition)
        keys = sorted(map(str, keys))
        return keys

    def fit_model(self, glycopeptide_chromatograms):
        model = self._model_class(
            glycopeptide_chromatograms, self.factors,
            replicate_key_attr=self.replicate_key_attr,
            use_retention_time_normalization=self.use_retention_time_normalization)
        model.fit()
        return model

    def fit(self):
        self.log("Fitting Joint Model")
        model = self.fit_model(self.glycopeptide_chromatograms)
        self.log("R^2: %0.3f, MSE: %0.3f" % (model.R2(), model.mse))
        if self.refit_filter != 0.0:
            self.log("Filtering Training Data")
            filtered_cases = [
                case for case in self.glycopeptide_chromatograms
                if model.score(case) > self.refit_filter
            ]
            self.log("Re-fitting After Filtering")
            model = self.fit_model(filtered_cases)
            self.log("R^2: %0.3f, MSE: %0.3f" % (model.R2(), model.mse))
        self.log('\n' + model.summary())
        self.joint_model = model
        factors = sorted(self.factors)
        self.log("Measuring Single Monosaccharide Deltas, Median and MAD")
        for key in factors:
            deltas = self._deltas_for(key)
            self.delta_by_factor[key] = deltas
            self.log("%s:   %0.3f   %0.3f" %
                     (key,
                      np.median(deltas),
                      np.median(np.abs(deltas - np.median(deltas)), )))
        for key, members in self.by_peptide.items():
            distinct_members = set(str(m.structure) for m in members)
            self.log("Fitting Model For %s (%d observations, %d distinct)" % (key, len(members), len(distinct_members)))
            if len(distinct_members) <= max(len(self.factors), self.minimum_observations_for_specific_model):
                self.log("Too few distinct observations for %s" % (key, ))
                continue
            model = self.fit_model(members)
            self.log("R^2: %0.3f, MSE: %0.3f" % (model.R2(), model.mse))
            if self.refit_filter != 0.0:
                self.log("Filtering Training Data")
                filtered_cases = [
                    case for case in members
                    if model.score(case) > self.refit_filter
                ]
                self.log("Re-fitting After Filtering")
                model = self.fit_model(filtered_cases)
                self.log("R^2: %0.3f, MSE: %0.3f" % (model.R2(), model.mse))
            self.log('\n' + model.summary())
            self.peptide_specific_models[key] = model
            joint_perf = np.mean(map(self.joint_model.score, members))
            spec_perf = np.mean(map(model.score, members))
            self.log("Mean Peptide Model Score: %0.3f" % (spec_perf, ))
            self.log("Mean Joint Model Score:   %0.3f" % (joint_perf, ))

    def evaluate_training(self):
        for key, group in self.by_peptide.items():
            self.log("Evaluating %s" % key)
            for obs in group:
                model = self._model_for(obs)
                score = model.score(obs)
                pred = model.predict(obs)
                delta = model._get_apex_time(obs) - pred
                obs.annotations['score'] = score
                obs.annotations['predicted_apex_time'] = pred
                obs.annotations['delta_apex_time'] = delta

    def evaluate(self, observations):
        for obs in observations:
            model = self._model_for(obs)
            score = model.score(obs)
            pred = model.predict(obs)
            delta = model._get_apex_time(obs) - pred
            obs.annotations['score'] = score
            obs.annotations['predicted_apex_time'] = pred
            obs.annotations['delta_apex_time'] = delta

    def _model_for(self, observation):
        if self.prefer_joint_model:
            return self.joint_model
        key = glycopeptidepy.parse(str(observation.structure)).deglycosylate()
        model = self.peptide_specific_models.get(key, self.joint_model)
        return model

    def predict(self, observation):
        model = self._model_for(observation)
        return model.predict(observation)

    def score(self, observation):
        model = self._model_for(observation)
        return model.score(observation)

    def run(self):
        self.fit()
        self.evaluate_training()
        if self.test_chromatograms:
            self.evaluate(self.test_chromatograms)

    def write(self, path):
        from glycresoft.output.report.base import render_plot
        from glycresoft.plotting.base import figax
        if not os.path.exists(path):
            os.makedirs(path)
        elif not os.path.isdir(path):
            raise IOError("Expected a path to a directory, %s is a file!" % (path, ))
        pjoin = os.path.join
        self.log("Writing scored chromatograms")
        with csv_stream(open(pjoin(path, "scored_chromatograms.csv"), 'wb')) as fh:
            GlycopeptideChromatogramProxy.to_csv(self.glycopeptide_chromatograms, fh)
        if self.test_chromatograms:
            with csv_stream(open(pjoin(path, "test_chromatograms.csv"), 'wb')) as fh:
                GlycopeptideChromatogramProxy.to_csv(self.test_chromatograms, fh)
        self.log("Writing joint model descriptors")
        with csv_stream(open(pjoin(path, "joint_model_parameters.csv"), 'wb')) as fh:
            self.joint_model.to_csv(fh)
        with open(pjoin(path, "joint_model_predplot.png"), 'wb') as fh:
            ax = figax()
            self.joint_model.prediction_plot(ax=ax)
            fh.write(render_plot(ax, dpi=160.0).getvalue())

        if self.use_retention_time_normalization:
            with open(pjoin(path, "retention_time_normalization.png"), 'wb') as fh:
                ax = figax()
                self.joint_model.run_normalizer.plot(ax=ax)
                ax.set_title("Cross-Run RT Correction")
                fh.write(render_plot(ax, dpi=160.0).getvalue())

        for key, model in self.peptide_specific_models.items():
            self.log("Writing %s model descriptors" % (key, ))
            with csv_stream(open(pjoin(path, "%s_model_parameters.csv" % (key, )), 'wb')) as fh:
                model.to_csv(fh)
            with open(pjoin(path, "%s_model_predplot.png" % (key, )), 'wb') as fh:
                ax = figax()
                model.prediction_plot(ax=ax)
                ax.set_title(key)
                fh.write(render_plot(ax, dpi=160.0).getvalue())

        for factor, deltas in self.delta_by_factor.items():
            with open(pjoin(path, '%s_delta_hist.png' % (factor.replace("@", ""), )), 'wb') as fh:
                ax = figax()

                ax.hist(deltas, bins='auto', density=False, ec='black', alpha=0.5)
                ax.set_title(factor)
                ax.figure.text(0.75, 0.8, 'Median: %0.3f' % np.median(deltas), ha='center')
                ax.set_xlabel("RT Shift (Min)")
                ax.set_ylabel("Count")
                ax2 = ax.twinx()
                if len(deltas) > 1:
                    x = np.linspace(deltas.min() + -0.1, deltas.max() + 0.1)
                    m = gaussian_kde(deltas)
                    ax2.fill_between(x, 0, m.pdf(x), alpha=0.5, color='green')
                    ax2.set_ylabel("Density")
                    ax2.set_ylim(0, ax2.get_ylim()[1])
                else:
                    m = None
                fh.write(render_plot(ax, dpi=160.0).getvalue())

        self.log("Saving models")
        with open(pjoin(path, "model.pkl"), 'wb') as fh:
            pickle.dump(self, fh, -1)



from collections import defaultdict

from glycopeptidepy.structure import sequence

from glycresoft.chromatogram_tree.relation_graph import (
    ChromatogramGraph, ChromatogramGraphEdge, ChromatogramGraphNode)


class GlycopeptideChromatogramGraphNode(ChromatogramGraphNode):
    def __init__(self, chromatogram, index, edges=None):
        super(GlycopeptideChromatogramGraphNode, self).__init__(chromatogram, index, edges)
        self.backbone = None
        if chromatogram.composition:
            if hasattr(chromatogram, 'entity'):
                entity = chromatogram.entity
                self.backbone = entity.get_sequence(include_glycan=False)


class GlycopeptideChromatogramGraph(ChromatogramGraph):
    def __init__(self, chromatograms):
        self.sequence_map = defaultdict(list)
        super(GlycopeptideChromatogramGraph, self).__init__(chromatograms)

    def _construct_graph_nodes(self, chromatograms):
        nodes = []
        for i, chroma in enumerate(chromatograms):
            node = (GlycopeptideChromatogramGraphNode(chroma, i))
            nodes.append(node)
            if node.chromatogram.composition:
                self.enqueue_seed(node)
                self.assignment_map[node.chromatogram.composition] = node
                if node.backbone is not None:
                    self.sequence_map[node.backbone].append(node)
        return nodes

    def find_edges(self, node, query_width=2., transitions=None, **kwargs):
        super(GlycopeptideChromatogramGraph, self).find_edges(node, query_width, transitions)
        if node.backbone is not None:
            for other_node in self.sequence_map[node.backbone]:
                ppm_error = 0
                rt_error = node.center - other_node.center
                ChromatogramGraphEdge(node, other_node, 'backbone', ppm_error, rt_error)



import unittest
import tempfile

from glycresoft.database.builder.glycan import glycan_source


FILE_SOURCE = '''
{Hex:3; HexNAc:2}  N-Glycan  O-Glycan
{Fuc:1; Hex:3; HexNAc:2}  N-Glycan
'''


GLYCONNECT_SOURCE = '''
HexNAc:2 Hex:3  N-Glycan  O-Glycan
dHex:1 Hex:3 HexNAc:2  N-Glycan
'''


BYONIC_SOURCE = '''
HexNAc(2)Hex(3)  N-Glycan  O-Glycan
dHex(1)Hex(3)HexNAc(2)  N-Glycan
'''

class GlycanSourceTests(unittest.TestCase):

    def setup_tempfile(self):
        file_name = tempfile.mktemp() + ".txt"
        open(file_name, 'w').write(FILE_SOURCE)
        return file_name

    def clear_file(self, path):
        open(path, 'wb')

    def test_run(self):
        file_name = self.setup_tempfile()
        builder = glycan_source.TextFileGlycanHypothesisSerializer(
            file_name, file_name + '.db')
        builder.start()
        inst = builder.query(glycan_source.DBGlycanComposition).filter(
            glycan_source.DBGlycanComposition.hypothesis_id == builder.hypothesis_id,
            glycan_source.DBGlycanComposition.composition == "{Hex:3; HexNAc:2}").one()
        self.assertAlmostEqual(inst.calculated_mass, 910.32777, 3)
        self.assertTrue("N-Glycan" in inst.structure_classes)
        builder.engine.dispose()
        self.clear_file(file_name + '.db')
        self.clear_file(file_name)

    def test_run_reduced(self):
        file_name = self.setup_tempfile()
        builder = glycan_source.TextFileGlycanHypothesisSerializer(
            file_name, file_name + '.db', reduction="H2")
        builder.start()
        inst = builder.query(glycan_source.DBGlycanComposition).filter(
            glycan_source.DBGlycanComposition.hypothesis_id == builder.hypothesis_id,
            glycan_source.DBGlycanComposition.composition == "{Hex:3; HexNAc:2}$H2").one()
        self.assertAlmostEqual(inst.calculated_mass, 912.3434, 3)
        builder.engine.dispose()
        self.clear_file(file_name + '.db')
        self.clear_file(file_name)

    def test_run_permethylated(self):
        file_name = self.setup_tempfile()
        builder = glycan_source.TextFileGlycanHypothesisSerializer(
            file_name, file_name + '.db', reduction="H2", derivatization='methyl')
        builder.start()
        inst = builder.query(glycan_source.DBGlycanComposition).filter(
            glycan_source.DBGlycanComposition.hypothesis_id == builder.hypothesis_id,
            glycan_source.DBGlycanComposition.composition == "{Hex^Me:3; HexNAc^Me:2}$C1H4").one()
        self.assertAlmostEqual(inst.calculated_mass, 1164.6251311968801, 3)
        builder.engine.dispose()
        self.clear_file(file_name + '.db')
        self.clear_file(file_name)


class GlyconnectGlycanSourceTests(unittest.TestCase):
    def setup_tempfile(self):
        file_name = tempfile.mktemp() + ".txt"
        open(file_name, 'w').write(GLYCONNECT_SOURCE)
        return file_name

    def clear_file(self, path):
        open(path, 'wb')

    def test_run(self):
        file_name = self.setup_tempfile()
        builder = glycan_source.TextFileGlycanHypothesisSerializer(
            file_name, file_name + '.db')
        builder.start()
        inst = builder.query(glycan_source.DBGlycanComposition).filter(
            glycan_source.DBGlycanComposition.hypothesis_id == builder.hypothesis_id,
            glycan_source.DBGlycanComposition.composition == "{Hex:3; HexNAc:2}").one()
        self.assertAlmostEqual(inst.calculated_mass, 910.32777, 3)
        self.assertTrue("N-Glycan" in inst.structure_classes)
        builder.engine.dispose()
        self.clear_file(file_name + '.db')
        self.clear_file(file_name)


class ByonicGlycanSourceTests(unittest.TestCase):
    def setup_tempfile(self):
        file_name = tempfile.mktemp() + ".txt"
        open(file_name, 'w').write(BYONIC_SOURCE)
        return file_name

    def clear_file(self, path):
        open(path, 'wb')

    def test_run(self):
        file_name = self.setup_tempfile()
        builder = glycan_source.TextFileGlycanHypothesisSerializer(
            file_name, file_name + '.db')
        builder.start()
        inst = builder.query(glycan_source.DBGlycanComposition).filter(
            glycan_source.DBGlycanComposition.hypothesis_id == builder.hypothesis_id,
            glycan_source.DBGlycanComposition.composition == "{Hex:3; HexNAc:2}").one()
        self.assertAlmostEqual(inst.calculated_mass, 910.32777, 3)
        self.assertTrue("N-Glycan" in inst.structure_classes)
        builder.engine.dispose()
        self.clear_file(file_name + '.db')
        self.clear_file(file_name)


if __name__ == '__main__':
    unittest.main()



from operator import mul

from typing import Generic, TypeVar, Optional, OrderedDict, Dict, Any

try:
    reduce
except NameError:
    from functools import reduce

import numpy as np

from .shape_fitter import ChromatogramShapeModel
from .spacing_fitter import ChromatogramSpacingModel
from .charge_state import UniformChargeStateScoringModel
from .isotopic_fit import IsotopicPatternConsistencyModel
from .base import symbolic_composition, epsilon, ScoringFeatureBase
from .utils import logitsum, prod

from glycresoft.chromatogram_tree import ChromatogramInterface, Chromatogram


class ScorerBase(object):
    feature_set: OrderedDict[str, ScoringFeatureBase]
    configuration: Dict[str, Any]

    def __init__(self, *args, **kwargs):
        self.feature_set = OrderedDict()
        self.configuration = dict()

    def __getitem__(self, k):
        return self.feature_set[k]

    def __contains__(self, k):
        return k in self.feature_set

    def __iter__(self):
        return iter(self.feature_set)

    def __len__(self):
        return len(self.feature_set)

    def configure(self, analysis_info):
        for feature in self.features():
            config_data = feature.configure(analysis_info)
            self.configuration[feature.get_feature_name()] = config_data

    def add_feature(self, scoring_feature):
        if scoring_feature is None:
            return
        feature_type = scoring_feature.get_feature_type()
        self.feature_set[feature_type] = scoring_feature

    def features(self):
        for feature in self.feature_set.values():
            yield feature

    def compute_scores(self, chromatogram: ChromatogramInterface) -> 'ChromatogramScoreSet':
        raise NotImplementedError()

    def logitscore(self, chromatogram: ChromatogramInterface) -> float:
        score = logitsum(self.compute_scores(chromatogram))
        return score

    def score(self, chromatogram: ChromatogramInterface) -> float:
        score = reduce(mul, self.compute_scores(chromatogram), 1.0)
        return score

    def accept(self, solution: 'ChromatogramSolution') -> bool:
        scored_features = solution.score_components()
        for feature in self.features():
            if feature.reject(scored_features, solution):
                return False
        return True

    def __repr__(self):
        template = "{self.__class__.__name__}({features})"
        features = ', '.join(["%r: %s" % kv for kv in self.feature_set.items()])
        return template.format(self=self, features=features)


class ChromatogramScoreSet(object):
    scores: OrderedDict[str, float]

    def __init__(self, scores):
        self.scores = OrderedDict(scores)

    def __iter__(self):
        return iter(self.scores.values())

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, str(dict(self.scores))[1:-1])

    def __getitem__(self, key):
        return self.scores[key]

    def __setitem__(self, key, value):
        self.scores[key] = value

    def __getattr__(self, key):
        if key == "scores":
            raise AttributeError(key)
        try:
            return self.scores[key]
        except KeyError:
            raise AttributeError(key)

    def keys(self):
        return self.scores.keys()

    def values(self):
        return self.scores.values()

    def items(self):
        return self.scores.items()

    def __mul__(self, i):
        new = self.__class__(self.scores)
        for feature in self.keys():
            new[feature] *= i
        return new

    def __div__(self, i):
        new = self.__class__(self.scores)
        for feature in self.keys():
            new[feature] /= i
        return new

    def __add__(self, other):
        new = self.__class__(self.scores)
        for key, value in other.items():
            try:
                new[key] += value
            except KeyError:
                new[key] = value
        return new

    def product(self) -> float:
        return prod(*self)

    def logitsum(self) -> float:
        return logitsum(self)


class DummyScorer(ScorerBase):
    def __init__(*args, **kwargs):
        pass

    def score(self, *args, **kwargs):
        return 1.0

    @property
    def line_test(self):
        return 0.0

    @property
    def spacing_fit(self):
        return 0.0

    @property
    def mean_fit(self):
        return 0.0

    def compute_scores(self, chromatogram):
        line_score = 1 - epsilon
        isotopic_fit = 1 - epsilon
        spacing_fit = 1 - epsilon
        charge_count = 1 - epsilon
        return ChromatogramScoreSet([
            ("line_score", line_score), ("isotopic_fit", isotopic_fit),
            ("spacing_fit", spacing_fit), ("charge_count", charge_count)
        ])

    def logitscore(self, chromatogram):
        score = logitsum(self.compute_scores(chromatogram))
        return score


class ChromatogramScorer(ScorerBase):
    def __init__(self, shape_fitter_model=ChromatogramShapeModel,
                 isotopic_fitter_model=IsotopicPatternConsistencyModel,
                 charge_scoring_model=UniformChargeStateScoringModel,
                 spacing_fitter_model=ChromatogramSpacingModel,
                 mass_shift_scoring_model=None, *models):
        super(ChromatogramScorer, self).__init__()
        self.add_feature(shape_fitter_model())
        self.add_feature(isotopic_fitter_model())
        self.add_feature(spacing_fitter_model())
        self.add_feature(charge_scoring_model())
        if mass_shift_scoring_model is not None:
            self.add_feature(mass_shift_scoring_model())
        for model in models:
            if model is not None:
                self.add_feature(model())

    def compute_scores(self, chromatogram):
        scores = []
        for model in self.features():
            scores.append((model.get_feature_type(), model.score(chromatogram)))
        return ChromatogramScoreSet(scores)

    def clone(self):
        dup = self.__class__()
        for feature in self.features():
            dup.add_feature(feature)
        return dup


class ModelAveragingScorer(ScorerBase):
    def __init__(self, models, weights=None):
        if weights is None:
            weights = [1.0 for i in range(len(models))]
        self.models = models
        self.weights = weights

        self.prepare_weights()

    def prepare_weights(self):
        a = np.array(self.weights)
        a /= a.sum()
        self.weights = a

    def compute_scores(self, chromatogram):
        score_set = ChromatogramScoreSet({})
        weights = 0
        for model, weight in zip(self.models, self.weights):
            score = model.compute_scores(chromatogram)
            score_set += (score * weight)
            weights += weight
        return score_set / weights

    def clone(self):
        return self.__class__(list(self.models), list(self.weights))


class CompositionDispatchScorer(ScorerBase):
    def __init__(self, rule_model_map, default_model=None):
        if default_model is None:
            default_model = ChromatogramScorer()
        self.rule_model_map = rule_model_map
        self.default_model = default_model

    def clone(self):
        return self.__class__({k: v.clone() for k, v in self.rule_model_map.items()}, self.default_model.clone())

    def get_composition(self, obj):
        if obj.composition is not None:
            if obj.glycan_composition is not None:
                return symbolic_composition(obj)
        return None

    def find_model(self, composition):
        if composition is None:
            return self.default_model
        for rule in self.rule_model_map:
            if rule(composition):
                return self.rule_model_map[rule]

        return self.default_model

    def compute_scores(self, chromatogram):
        composition = self.get_composition(chromatogram)
        model = self.find_model(composition)
        return model.compute_scores(chromatogram)

    def accept(self, solution):
        composition = self.get_composition(solution)
        model = self.find_model(composition)
        return model.accept(solution)


C = TypeVar("C", bound=ChromatogramInterface)


class ChromatogramSolution(Generic[C]):
    _temp_score = 0.0

    chromatogram: C
    scorer: Optional[ScorerBase]
    score: Optional[float]
    internal_score: Optional[float]
    _temp_score: float
    score_set: Optional[ChromatogramScoreSet]

    def __init__(self, chromatogram, score=None, scorer=ChromatogramScorer(), internal_score=None,
                 score_set=None):
        if internal_score is None:
            internal_score = score
        if isinstance(chromatogram, ChromatogramSolution):
            chromatogram = chromatogram.get_chromatogram()
        self.chromatogram = chromatogram
        self.scorer = scorer
        self.internal_score = internal_score
        self.score = score

        if score is None:
            self.compute_score()

        self.score_set = score_set

    def __getattr__(self, name):
        return getattr(object.__getattribute__(self, "chromatogram"), name)

    def __len__(self):
        return len(self.chromatogram)

    def __iter__(self):
        return iter(self.chromatogram)

    def __eq__(self, other):
        return self.chromatogram.__eq__(other)

    def __hash__(self):
        return self.chromatogram.__hash__()

    def compute_score(self):
        try:
            self.internal_score = self.score = self.scorer.score(self.chromatogram)
        except ZeroDivisionError:
            self.internal_score = self.score = 0.0

    def score_components(self):
        if self.score_set is None:
            self.score_set = self.scorer.compute_scores(self.chromatogram)
        return self.score_set

    @property
    def logitscore(self) -> float:
        return self.scorer.logitscore(self.chromatogram)

    def get_chromatogram(self):
        return self.chromatogram

    def clone(self):
        data = self.chromatogram.clone()
        return self.__class__(data, self.score, self.scorer, self.internal_score)

    def __repr__(self):
        return "ChromatogramSolution(%s, %0.4f, %d, %0.4f)" % (
            self.chromatogram.composition, self.chromatogram.neutral_mass,
            self.chromatogram.n_charge_states, self.score)

    def __getitem__(self, i):
        return self.chromatogram[i]

    def __dir__(self):
        return list(set(('compute_score', 'score_components')) | set(dir(self.chromatogram)))

    def __eq__(self, other):
        return self.get_chromatogram() == other.get_chromatogram()

    def __ne__(self, other):
        return self.get_chromatogram() != other.get_chromatogram()


ChromatogramInterface.register(ChromatogramSolution)



from .linear_regression import (weighted_linear_regression_fit, prediction_interval, ransac)

from .structure import (
    ChromatogramProxy, GlycopeptideChromatogramProxy,
    CommonGlycopeptide, _get_apex_time, GlycoformAggregator)

from .model import (
    ElutionTimeFitter, AbundanceWeightedElutionTimeFitter,
    FactorElutionTimeFitter, AbundanceWeightedFactorElutionTimeFitter,
    PeptideFactorElutionTimeFitter, AbundanceWeightedPeptideFactorElutionTimeFitter,
    ElutionTimeModel, ModelBuildingPipeline as GlycopeptideElutionTimeModelBuildingPipeline,
    ModelEnsemble)

from .reviser import (
    ModelReviser, IntervalModelReviser, PeptideYUtilizationPreservingRevisionValidator,
    OxoniumIonRequiringUtilizationRevisionValidator, CompoundRevisionValidator,
    RevisionRule, ValidatingRevisionRule, MassShiftRule)

from .cross_run import (
    ReplicatedAbundanceWeightedPeptideFactorElutionTimeFitter, LinearRetentionTimeCorrector,
)

from .recalibrator import (CalibrationPoint, RecalibratingPredictor)

from .pipeline import GlycopeptideElutionTimeModeler



from glycresoft.database.builder.glycan import CombinatorialGlycanHypothesisSerializer
from glycresoft.database.prebuilt.utils import hypothesis_register, BuildBase
from io import StringIO


combinatorial_source = u'''Hex 3 10
HexNAc 2 9
Fuc 0 5
NeuAc 0 5
NeuGc 0 5

Fuc < HexNAc
HexNAc > (NeuAc + NeuGc) + 1'''

hypothesis_metadata = {
    "name": 'Combinatorial Mammalian N-Glycans',
    "hypothesis_type": 'glycan_composition'
}


# @hypothesis_register(hypothesis_metadata['name'])
class CombinatorialMammalianNGlycansBuilder(BuildBase):

    def get_hypothesis_metadata(self):
        return hypothesis_metadata

    def build(self, database_connection, **kwargs):
        kwargs.setdefault('hypothesis_name', self.hypothesis_metadata['name'])
        task = CombinatorialGlycanHypothesisSerializer(
            StringIO(combinatorial_source), database_connection,
            **kwargs)
        task.start()
        return task



import logging
import multiprocessing

import click

from glycresoft import version
from glycresoft.cli.logger_config import make_log_file_logger, LOG_FILE_MODE, LOG_LEVEL

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

logger = logging.getLogger("glycresoft")


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version.version)
@click.option("-l", "--log-file", type=click.Path(writable=True), required=False,
              default=None, help="Path to write the log file to.")
def cli(log_file=None):
    if log_file is not None:
        click.echo(f"Logging to {log_file}")
        make_log_file_logger(log_file_name=log_file, log_file_mode=LOG_FILE_MODE,
                             level=LOG_LEVEL, name='glycresoft')


class HiddenOption(click.Option):
    """
    HiddenOption -- absent from Help text.

    Supported in latest and greatest version of Click, but not old versions, so
    use generic 'cls=HiddenOption' to get the desired behavior.
    """

    @property
    def hidden(self):
        return True

    @hidden.setter
    def hidden(self, value):
        return

    def get_help_record(self, ctx):
        """
        Has "None" as its help record. All that is needed.
        """
        return


class DocumentableArgument(click.Argument):
    def __init__(self, *args, **kwargs):
        doc_help = kwargs.pop("doc_help", "")
        super(DocumentableArgument, self).__init__(*args, **kwargs)
        self.doc_help = doc_help


_option = click.option
_argument = click.argument


def option(*args, **kwargs):
    kwargs.setdefault("show_default", True)
    return _option(*args, **kwargs)


def argument(*args, **kwargs):
    kwargs['cls'] = DocumentableArgument
    a = _argument(*args, **kwargs)
    return a


click.option = option
click.argument = argument


processes_option = click.option(
    "-p", "--processes", 'processes', type=click.IntRange(1, multiprocessing.cpu_count()),
    default=min(multiprocessing.cpu_count(), 4), help=('Number of worker processes to use. Defaults to 4 '
                                                       'or the number of CPUs, whichever is lower'),
    # show_default=True
)



import sys
import traceback
import os
from setuptools import setup, Extension, find_packages

from distutils.command.build_ext import build_ext
from distutils.errors import (CCompilerError, DistutilsExecError,
                              DistutilsPlatformError)


def has_option(name):
    try:
        sys.argv.remove('--%s' % name)
        return True
    except ValueError:
        pass
    # allow passing all cmd line options also as environment variables
    env_val = os.getenv(name.upper().replace('-', '_'), 'false').lower()
    if env_val == "true":
        return True
    return False


include_diagnostics = has_option("include-diagnostics")
force_cythonize = has_option("force-cythonize")


def make_extensions():
    is_ci = bool(os.getenv("CI", ""))
    try:
        import numpy
    except ImportError:
        print("Installation requires `numpy`")
        raise
    macros = []
    try:
        from Cython.Build import cythonize
        cython_directives = {
            'embedsignature': True,
            "profile": include_diagnostics
        }
        if include_diagnostics:
            macros.append(("CYTHON_TRACE_NOGIL", "1"))
        if is_ci and include_diagnostics:
            cython_directives['linetrace'] = True
        extensions = cythonize(
            [
                Extension(
                    name="glycresoft._c.structure.fragment_match_map",
                    sources=["src/glycresoft/_c/structure/fragment_match_map.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.structure.intervals",
                    sources=["src/glycresoft/_c/structure/intervals.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.scoring.shape_fitter",
                    sources=["src/glycresoft/_c/scoring/shape_fitter.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.chromatogram_tree.mass_shift",
                    sources=["src/glycresoft/_c/chromatogram_tree/mass_shift.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.chromatogram_tree.index",
                    sources=["src/glycresoft/_c/chromatogram_tree/index.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.tandem.core_search",
                    sources=["src/glycresoft/_c/tandem/core_search.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.database.mass_collection",
                    sources=["src/glycresoft/_c/database/mass_collection.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.tandem.tandem_scoring_helpers",
                    libraries=["npymath"],
                    library_dirs=[os.path.join(os.path.dirname(numpy.get_include()), "lib")],
                    sources=["src/glycresoft/_c/tandem/tandem_scoring_helpers.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.tandem.spectrum_match",
                    sources=["src/glycresoft/_c/tandem/spectrum_match.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.composition_network.graph",
                    sources=["src/glycresoft/_c/composition_network/graph.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.composition_distribution_model.utils",
                    sources=["src/glycresoft/_c/composition_distribution_model/utils.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.structure.lru", sources=["src/glycresoft/_c/structure/lru.pyx"]
                ),
                Extension(
                    name="glycresoft._c.tandem.target_decoy",
                    sources=["src/glycresoft/_c/tandem/target_decoy.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.structure.structure_loader",
                    sources=["src/glycresoft/_c/structure/structure_loader.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.tandem.oxonium_ions",
                    sources=["src/glycresoft/_c/tandem/oxonium_ions.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.structure.probability",
                    sources=["src/glycresoft/_c/structure/probability.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.tandem.peptide_graph",
                    sources=["src/glycresoft/_c/tandem/peptide_graph.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
                Extension(
                    name="glycresoft._c.scoring.elution_time_grouping",
                    sources=["src/glycresoft/_c/scoring/elution_time_grouping.pyx"],
                    include_dirs=[numpy.get_include()],
                ),
            ],
            compiler_directives=cython_directives,
            force=force_cythonize,
        )
    except ImportError as err:
        print(err)
        extensions = [
            Extension(
                name="glycresoft._c.structure.fragment_match_map",
                sources=["src/glycresoft/_c/structure/fragment_match_map.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                name="glycresoft._c.structure.intervals",
                sources=["src/glycresoft/_c/structure/intervals.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                name="glycresoft._c.scoring.shape_fitter",
                sources=["src/glycresoft/_c/scoring/shape_fitter.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                name="glycresoft._c.chromatogram_tree.mass_shift",
                sources=["src/glycresoft/_c/chromatogram_tree/mass_shift.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                name="glycresoft._c.chromatogram_tree.index",
                sources=["src/glycresoft/_c/chromatogram_tree/index.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                name="glycresoft._c.tandem.core_search",
                sources=["src/glycresoft/_c/tandem/core_search.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                name="glycresoft._c.database.mass_collection",
                sources=["src/glycresoft/_c/database/mass_collection.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                name="glycresoft._c.tandem.tandem_scoring_helpers",
                libraries=["npymath"],
                library_dirs=[os.path.join(os.path.dirname(numpy.get_include()), "lib")],
                sources=["src/glycresoft/_c/tandem/tandem_scoring_helpers.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                name="glycresoft._c.tandem.spectrum_match",
                sources=["src/glycresoft/_c/tandem/spectrum_match.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                name="glycresoft._c.composition_network.graph",
                sources=["src/glycresoft/_c/composition_network/graph.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                name="glycresoft._c.composition_distribution_model.utils",
                sources=["src/glycresoft/_c/composition_distribution_model/utils.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(name="glycresoft._c.structure.lru", sources=["src/glycresoft/_c/structure/lru.c"]),
            Extension(
                name="glycresoft._c.tandem.target_decoy",
                sources=["src/glycresoft/_c/tandem/target_decoy.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                name="glycresoft._c.structure.structure_loader",
                sources=["src/glycresoft/_c/structure/structure_loader.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                name="glycresoft._c.tandem.oxonium_ions",
                sources=["src/glycresoft/_c/tandem/oxonium_ions.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                name="glycresoft._c.structure.probability",
                sources=["src/glycresoft/_c/structure/probability.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                name="glycresoft._c.tandem.peptide_graph",
                sources=["src/glycresoft/_c/tandem/peptide_graph.c"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                name="glycresoft._c.scoring.elution_time_grouping",
                sources=["src/glycresoft/_c/scoring/elution_time_grouping.c"],
                include_dirs=[numpy.get_include()],
            ),
        ]
    return extensions


ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)
if sys.platform == 'win32':
    # 2.6's distutils.msvc9compiler can raise an IOError when failing to
    # find the compiler
    ext_errors += (IOError,)


class BuildFailed(Exception):

    def __init__(self):
        self.cause = sys.exc_info()[1]  # work around py 2/3 different syntax

    def __str__(self):
        return str(self.cause)


class ve_build_ext(build_ext):
    # This class allows C extension building to fail.

    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            traceback.print_exc()
            raise BuildFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except ext_errors:
            traceback.print_exc()
            raise BuildFailed()
        except ValueError:
            # this can happen on Windows 64 bit, see Python issue 7511
            traceback.print_exc()
            if "'path'" in str(sys.exc_info()[1]):  # works with both py 2/3
                raise BuildFailed()
            raise


cmdclass = {}

cmdclass['build_ext'] = ve_build_ext


def status_msgs(*msgs):
    print('*' * 75)
    for msg in msgs:
        print(msg)
    print('*' * 75)


with open("src/glycresoft/version.py") as version_file:
    version = None
    for line in version_file.readlines():
        if "version = " in line:
            version = line.split(" = ")[1].replace("\"", "").strip()
            print("Version is: %r" % (version,))
            break
    else:
        print("Cannot determine version")


requirements = []
with open("requirements.txt") as requirements_file:
    requirements.extend(requirements_file.readlines())

try:
    with open("README.md") as readme_file:
        long_description = readme_file.read()
except Exception as e:
    print(e)
    long_description = ""


def run_setup(include_cext=True):

    setup(
        name="glycresoft",
        version=version,
        packages=find_packages(where="src"),
        package_dir={"": "src"},
        include_package_data=True,
        author=", ".join(["Joshua Klein"]),
        author_email="jaklein@bu.edu",
        description="Glycan and Glycopeptide Mass Spectrometry Database Search Tool",
        long_description=long_description,
        long_description_content_type="text/markdown",
        entry_points={
            "console_scripts": ["glycresoft = glycresoft.cli.__main__:main"],
        },
        extras_require={"fragmentation-modeling": ["glycopeptide-feature-learning"]},
        package_data={
            "glycresoft.models": ["src/glycresoft/models/data/*"],
            "glycresoft.database.prebuilt": ["src/glycresoft/database/prebuilt/data/*"],
            "glycresoft.output.report.glycan_lcms": ["src/glycresoft/output/report/glycan_lcms/*"],
            "glycresoft.output.report.glycopeptide_lcmsms": ["src/glycresoft/output/report/glycopeptide_lcmsms/*"],
        },
        ext_modules=make_extensions() if include_cext else None,
        cmdclass=cmdclass,
        install_requires=requirements,
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: Apache Software License",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
        ],
        zip_safe=False,
        python_requires=">3.8",
        project_urls={
            "Documentation": "https://mobiusklein.github.io/glycresoft",
            "Source Code": "https://github.com/mobiusklein/glycresoft",
            "Issue Tracker": "https://github.com/mobiusklein/glycresoft/issues",
        },
    )


try:
    run_setup(True)
except Exception as exc:
    print(exc)
    run_setup(False)

    status_msgs(
        "WARNING: The C extension could not be compiled, " +
        "speedups are not enabled.",
        "Plain-Python build succeeded."
    )



import os
import re
import sys
import uuid
import codecs
import logging
import warnings
import multiprocessing

from typing import Dict
from logging import FileHandler, LogRecord
from multiprocessing import current_process

from stat import S_IRUSR, S_IWUSR, S_IXUSR, S_IRGRP, S_IWGRP, S_IROTH, S_IWOTH

from ms_deisotope.task.log_utils import LogUtilsMixin as _LogUtilsMixin

from glycresoft import task
from glycresoft.config import config_file

logging_levels = {
    'CRITICAL': 50,
    'DEBUG': 10,
    'ERROR': 40,
    'INFO': 20,
    'NOTSET': 0,
    'WARN': 30,
    'WARNING': 30
}

LOG_FILE_NAME = "glycresoft-log"
LOG_FILE_MODE = 'w'
LOG_LEVEL = 'INFO'

user_config_data = config_file.get_configuration()
try:
    LOG_FILE_NAME = user_config_data['environment']['log_file_name']
except KeyError:
    pass
try:
    LOG_FILE_MODE = user_config_data['environment']['log_file_mode']
except KeyError:
    pass
try:
    LOG_LEVEL = user_config_data['environment']['log_level']
    LOG_LEVEL = str(LOG_LEVEL).upper()
except KeyError:
    pass


LOG_FILE_NAME = os.environ.get("GLYCRESOFT_LOG_FILE_NAME", LOG_FILE_NAME)
LOG_FILE_MODE = os.environ.get("GLYCRESOFT_LOG_FILE_MODE", LOG_FILE_MODE)
LOG_LEVEL = str(os.environ.get("GLYCRESOFT_LOG_LEVEL", LOG_LEVEL)).upper()


log_multiprocessing = False
try:
    log_multiprocessing = bool(user_config_data['environment']['log_multiprocessing'])
except KeyError:
    pass


log_multiprocessing = bool(int(os.environ.get(
    "GLYCRESOFT_LOG_MULTIPROCESSING", log_multiprocessing)))


LOGGING_CONFIGURED = False


class ProcessAwareFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        d: Dict[str, str] = record.__dict__
        try:
            d['filename'] = d['filename'].replace(".py", '').ljust(13)[:13]
        except KeyError:
            d['filename'] = ''
        try:
            if d['processName'] == "MainProcess":
                d['maybeproc'] = ''
            else:
                d['maybeproc'] = "| %s " % d['processName'].replace(
                    "Process", '')
        except KeyError:
            d['maybeproc'] = ''
        return super(ProcessAwareFormatter, self).format(record)


class LevelAwareColoredLogFormatter(ProcessAwareFormatter):
    try:
        from colorama import Fore, Style, init
        init()
        GREY = Fore.WHITE
        BLUE = Fore.BLUE
        GREEN = Fore.GREEN
        YELLOW = Fore.YELLOW
        RED = Fore.RED
        BRIGHT = Style.BRIGHT
        DIM = Style.DIM
        BOLD_RED = Fore.RED + Style.BRIGHT
        RESET = Style.RESET_ALL
    except ImportError:
        GREY = ''
        BLUE = ''
        GREEN = ''
        YELLOW = ''
        RED = ''
        BRIGHT = ''
        DIM = ''
        BOLD_RED = ''
        RESET = ''

    def _colorize_field(self, fmt: str, field: str, color: str) -> str:
        return re.sub("(" + field + ")", color + r"\1" + self.RESET, fmt)

    def _patch_fmt(self, fmt: str, level_color: str) -> str:
        fmt = self._colorize_field(fmt, r"%\(asctime\)s", self.GREEN)
        fmt = self._colorize_field(fmt, r"%\(name\).*?s", self.BLUE)
        # fmt = self._colorize_field(fmt, r"%\(message\).*?s", self.GREY)
        if level_color:
            fmt = self._colorize_field(fmt, r"%\(levelname\).*?s", level_color)
        return fmt

    def __init__(self, fmt, level_color=None, **kwargs):
        fmt = self._patch_fmt(fmt, level_color=level_color)
        super().__init__(fmt, **kwargs)


class ColoringFormatter(logging.Formatter):
    level_to_color = {
        logging.INFO: LevelAwareColoredLogFormatter.GREEN,
        logging.DEBUG: LevelAwareColoredLogFormatter.GREY + LevelAwareColoredLogFormatter.DIM,
        logging.WARN: LevelAwareColoredLogFormatter.YELLOW + LevelAwareColoredLogFormatter.BRIGHT,
        logging.ERROR: LevelAwareColoredLogFormatter.BOLD_RED,
        logging.CRITICAL: LevelAwareColoredLogFormatter.BOLD_RED,
        logging.FATAL: LevelAwareColoredLogFormatter.RED + LevelAwareColoredLogFormatter.DIM,
    }

    _formatters: Dict[int, LevelAwareColoredLogFormatter]

    def __init__(self, fmt: str, **kwargs):
        self._formatters = {}
        for level, style in self.level_to_color.items():
            self._formatters[level] = LevelAwareColoredLogFormatter(
                fmt, level_color=style, **kwargs)

    def format(self, record: logging.LogRecord) -> str:
        fmtr = self._formatters[record.levelno]
        return fmtr.format(record)


DATE_FMT = "%H:%M:%S"


class _LogFilter(logging.Filter):
    """Filter out log messages from redundant logger helpers"""

    def filter(self, record: LogRecord) -> bool:
        if record.module == 'log_utils':
            return False
        return super().filter(record)


def get_status_formatter():
    status_terminal_format_string = '[%(asctime)s] %(levelname).1s | %(name)s | %(message)s'
    # status_terminal_format_string = "%(asctime)s %(name)s:%(levelname).1s - %(message)s"

    if sys.stderr.isatty():
        terminal_formatter = ColoringFormatter(
            status_terminal_format_string, datefmt=DATE_FMT)
    else:
        terminal_formatter = logging.Formatter(
            status_terminal_format_string, datefmt=DATE_FMT)
    return terminal_formatter


def make_status_logger(name="glycresoft.status"):
    status_logger = logging.getLogger(name)
    status_logger.propagate = False

    terminal_formatter = get_status_formatter()
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(terminal_formatter)
    status_logger.addHandler(handler)
    return status_logger


def get_main_process_formatter():
    terminal_format_string = (
        "[%(asctime)s] %(levelname).1s | %(name)s | %(filename)s:%(lineno)-4d%(maybeproc)s| %(message)s"
    )
    if sys.stderr.isatty():
        fmt = ColoringFormatter(
            terminal_format_string, datefmt=DATE_FMT)
    else:
        fmt = ProcessAwareFormatter(
            terminal_format_string, datefmt=DATE_FMT)
    return fmt


def make_main_process_logger(name=None, level="INFO"):
    fmt = get_main_process_formatter()
    handler = logging.StreamHandler()
    handler.setFormatter(fmt)
    handler.setLevel(level)
    logging.getLogger(name=name).addHandler(handler)


def make_log_file_logger(log_file_name, log_file_mode, name=None, level="INFO"):
    file_fmter = ProcessAwareFormatter(
        "[%(asctime)s] %(levelname).1s | %(name)s | %(filename)s:%(lineno)-4d%(maybeproc)s | %(message)s",
        DATE_FMT)
    flex_handler = FlexibleFileHandler(log_file_name, mode=log_file_mode)
    flex_handler.setFormatter(file_fmter)
    flex_handler.setLevel(level)
    logger = logging.getLogger(name=name)
    for handler in list(logger.handlers):
        if isinstance(handler, FlexibleFileHandler):
            logger.removeHandler(handler)
    logger.addHandler(flex_handler)


def configure_logging(level=None, log_file_name=None, log_file_mode=None):
    global LOGGING_CONFIGURED
    # If we've already called this, don't repeat it
    if LOGGING_CONFIGURED:
        return
    else:
        LOGGING_CONFIGURED = True
    if level is None:
        level = logging_levels.get(LOG_LEVEL, "INFO")
    if log_file_name is None:
        log_file_name = LOG_FILE_NAME
    if log_file_mode is None:
        log_file_mode = LOG_FILE_MODE

    if log_file_mode not in ("w", "a"):
        warnings.warn("File Logger configured with mode %r not applicable, using \"w\" instead" % (
            log_file_mode,))
        log_file_mode = "w"
    logging.getLogger().setLevel(level)

    log_filter = _LogFilter()

    logger_to_silence = logging.getLogger("ms_deisotope")
    logger_to_silence.setLevel(logging.INFO)
    logger_to_silence.addFilter(log_filter)
    logging.captureWarnings(True)

    logger = logging.getLogger("glycresoft")
    # We probably only need to register the base class but let's play it safe
    _LogUtilsMixin.log_with_logger(logger)
    task.LoggingMixin.log_with_logger(logger)
    task.TaskBase.log_with_logger(logger)

    make_status_logger()

    if current_process().name == "MainProcess":
        make_main_process_logger(level=level)

        if log_multiprocessing:
            fmt = get_main_process_formatter()
            multilogger = multiprocessing.get_logger()
            handler = logging.StreamHandler()
            handler.setFormatter(fmt)
            handler.setLevel(level)
            multilogger.addHandler(handler)

        if log_file_name:
            make_log_file_logger(log_file_name, log_file_mode, name='glycresoft', level=level)

    warner = logging.getLogger('py.warnings')
    warner.setLevel("CRITICAL")


permission_mask = S_IRUSR & S_IWUSR & S_IXUSR & S_IRGRP & S_IWGRP & S_IROTH & S_IWOTH


class FlexibleFileHandler(FileHandler):
    def _get_available_file(self):
        basename = current_name = self.baseFilename
        suffix = 0
        while suffix < 2 ** 16:
            if not os.path.exists(current_name):
                break
            elif os.access(current_name, os.W_OK):
                break
            else:
                suffix += 1
                current_name = "%s.%s" % (basename, suffix)
        if suffix < 2 ** 16:
            return current_name
        else:
            return "%s.%s" % (basename, uuid.uuid4().hex)

    def _open(self):
        """
        Open the current base file with the (original) mode and encoding.
        Return the resulting stream.
        """
        stream = LazyFile(self.baseFilename, self.mode, self.encoding)
        return stream


class LazyFile(object):
    def __init__(self, name, mode='r', encoding=None):
        self.name = name
        self.mode = mode
        self.encoding = encoding
        self._file = None

    def _open(self):
        file_name = self._get_available_file()
        if self.encoding is None:
            stream = open(file_name, self.mode)
        else:
            try:
                stream = codecs.open(file_name, self.mode, self.encoding)
            except LookupError:
                print(f"!! Failed to look up encoding {self.encoding}, falling back to UTF-8")
                stream = open(file_name, self.mode, encoding='utf8')
        self.name = file_name
        self._file = stream
        return self._file

    def read(self, n=None):
        if self._file is None:
            self._open()
        return self._file.read(n)

    def write(self, t):
        if self._file is None:
            self._open()
        return self._file.write(t)

    def close(self):
        if self._file is not None:
            self._file.close()
        return None

    def flush(self):
        if self._file is None:
            return
        return self._file.flush()

    def _get_available_file(self):
        basename = current_name = self.name
        suffix = 0
        while suffix < 2 ** 16:
            if not os.path.exists(current_name):
                break
            elif os.access(current_name, os.W_OK):
                break
            else:
                suffix += 1
                current_name = "%s.%s" % (basename, suffix)
        if suffix < 2 ** 16:
            return current_name
        else:
            return "%s.%s" % (basename, uuid.uuid4().hex)



import os
import sys
import traceback
import platform
import logging

from multiprocessing import freeze_support, set_start_method

import click

from glycresoft.cli import (
    base, build_db, tools, mzml, analyze, config,
    export)

try:
    from glycresoft_app.cli import server
except ImportError:
    pass


def info(type, value, tb):
    if not sys.stderr.isatty():
        click.secho("Running interactively, not starting debugger", fg='yellow')
        sys.__excepthook__(type, value, tb)
    else:
        import pdb
        traceback.print_exception(type, value, tb)
        pdb.post_mortem(tb)


def set_breakpoint_hook():
    try:
        import ipdb
        sys.breakpointhook = ipdb.set_trace
    except ImportError:
        pass


def main():
    freeze_support()
    try:
        if platform.system() == 'Windows' or platform.system() == "Darwin":
            set_start_method("spawn")
        else:
            set_start_method("forkserver")
    except Exception as err:
        logging.getLogger().error("Failed to set multiprocessing start method: %s", err, exc_info=True)

    if os.getenv("GLYCRESOFTDEBUG"):
        sys.excepthook = info
        click.secho("Running glycresoft with debugger", fg='yellow')
    if os.getenv("GLYCRESOFTPROFILING"):
        import cProfile
        click.secho("Running glycresoft with profiler", fg='yellow')
        profiler = cProfile.Profile()
        profiler.runcall(base.cli.main, standalone_mode=False)
        profiler.dump_stats('glycresoft_performance.profile')
    else:
        base.cli.main(standalone_mode=True)


if __name__ == '__main__':
    main()



def slurp(session, model, ids, flatten=True):
    if flatten:
        ids = [j for i in ids for j in i]
    total = len(ids)
    last = 0
    step = 100
    results = []
    while last < total:
        results.extend(session.query(model).filter(
            model.id.in_(ids[last:last + step])))
        last += step
    return results



import operator

import matplotlib
from matplotlib import font_manager
from matplotlib import pyplot as plt
from matplotlib import patches as mpatches
from matplotlib.textpath import TextPath
from matplotlib.transforms import IdentityTransform, Affine2D

import numpy as np

from .svg_utils import ET, BytesIO, IDMapper
from .colors import lighten, darken, get_color

from glycopeptidepy import PeptideSequence, enzyme


font_options = font_manager.FontProperties(family='monospace')


def span_overlap(a, b):
    if a.end_position == b.start_position or b.end_position == a.start_position:
        return False
    return (a.spans(b.start_position + 1) or a.spans(b.end_position) or
            b.spans(a.start_position + 1) or b.spans(a.end_position))


def layout_layers(gpms, sort_key=operator.attrgetter("ms2_score")):
    '''
    Produce a non-overlapping stacked layout of individual peptide-like
    identifications across a protein sequence.
    '''
    layers = [[]]
    gpms.sort(key=sort_key, reverse=True)
    for gpm in gpms:
        placed = False
        for layer in layers:
            collision = False
            for member in layer:
                if span_overlap(gpm, member):
                    collision = True
                    break
            if not collision:
                layer.append(gpm)
                placed = True
                break
        if not placed:
            layers.append([gpm])
    # import IPython
    # IPython.embed()
    return layers


def ndigits(x):
    digits = 0
    while x > 0:
        x //= 10
        digits += 1
    return digits


class GlycoformLayout(object):
    def __init__(self, protein, glycopeptides, scale_factor=1.0, ax=None, row_width=50,
                 sort_key=operator.attrgetter('ms2_score'), **kwargs):
        if ax is None:
            figure, ax = plt.subplots(1, 1)
        self.protein = protein
        self.sort_key = sort_key
        layers = self.layout_layers(glycopeptides)
        for layer in layers:
            layer.sort(key=lambda x: x.start_position)
        self.layers = layers
        self.id_mapper = IDMapper()
        self.ax = ax
        self.row_width = min(row_width, len(protein))
        self.options = kwargs
        self.layer_height = 0.56 * scale_factor
        self.y_step = (self.layer_height + 0.15) * -scale_factor

        self.cur_y = -3
        self.cur_position = 0

        self.mod_text_x_offset = 0.50 * scale_factor
        self.sequence_font_size = 6. * scale_factor
        self.mod_font_size = 2.08 * scale_factor
        self.mod_text_y_offset = 0.1 * scale_factor
        self.mod_width = 0.5 * scale_factor
        self.mod_x_offset = 0.60 * scale_factor
        self.total_length = len(protein.protein_sequence or '')
        self.protein_pad = -0.365 * scale_factor
        self.peptide_pad = self.protein_pad * (1.2)
        self.peptide_end_pad = 0.35 * scale_factor

        self.glycosites = set(protein.n_glycan_sequon_sites)

    def layout_layers(self, glycopeptides):
        layers = [[]]
        glycopeptides = list(glycopeptides)
        glycopeptides.sort(key=self.sort_key, reverse=True)
        for gpm in glycopeptides:
            placed = False
            for layer in layers:
                collision = False
                for member in layer:
                    if span_overlap(gpm, member):
                        collision = True
                        break
                if not collision:
                    layer.append(gpm)
                    placed = True
                    break
            if not placed:
                layers.append([gpm])
        return layers

    def draw_protein_main_sequence(self, current_position):
        next_row = current_position + self.row_width
        i = -1
        offset = current_position + 1
        digits = ndigits(offset)
        i -= (digits - 1) * 0.5
        text_path = TextPath(
            (self.protein_pad + i, self.layer_height + .2 + self.cur_y),
            str(current_position + 1), size=self.sequence_font_size / 7.5,
            prop=font_options)
        patch = mpatches.PathPatch(text_path, facecolor='grey', edgecolor='grey', lw=0.04)
        self.ax.add_patch(patch)

        i = self.row_width
        text_path = TextPath(
            (self.protein_pad + i, self.layer_height + .2 + self.cur_y),
            str(next_row), size=self.sequence_font_size / 7.5,
            prop=font_options)
        patch = mpatches.PathPatch(text_path, facecolor='grey', edgecolor='grey', lw=0.04)
        self.ax.add_patch(patch)
        self._draw_main_sequence(current_position, next_row)

    def _draw_main_sequence(self, start, end):
        for i, aa in enumerate(self.protein.protein_sequence[start:end]):
            text_path = TextPath(
                (self.protein_pad + i, self.layer_height + .2 + self.cur_y),
                aa, size=self.sequence_font_size / 7.5, prop=font_options)
            color = 'red' if any(
                (((i + start) in self.glycosites),
                 ((i + start - 1) in self.glycosites),
                 ((i + start - 2) in self.glycosites))
            ) else 'black'
            patch = mpatches.PathPatch(text_path, facecolor=color, edgecolor=color, lw=0.04)
            self.ax.add_patch(patch)

    def next_row(self):
        if self.cur_position > len(self.protein):
            return False
        self.cur_y += self.y_step * 3
        self.cur_position += self.row_width
        if self.cur_position >= len(self.protein):
            return False
        return True

    def _pack_sequence_metadata(self, rect, gpm):
        self.id_mapper.add("glycopeptide-%d", rect, {
            "sequence": str(gpm.structure),
            "start-position": gpm.start_position,
            "end-position": gpm.end_position,
            "ms2-score": gpm.ms2_score,
            "q-value": gpm.q_value,
            "record-id": gpm.id if hasattr(gpm, 'id') else None,
            "calculated-mass": gpm.structure.total_mass,
            "spectra-count": len(gpm.spectrum_matches)
        })

    def _get_sequence(self, gpm):
        try:
            return gpm.structure
        except AttributeError:
            return PeptideSequence(str(gpm))

    def draw_peptide_block(self, gpm, current_position, next_row):
        color, alpha = self._compute_sequence_color(gpm)

        interval_start = max(gpm.start_position - current_position, 0)
        interval_end = min(
            len(self._get_sequence(gpm)) + gpm.start_position - current_position,
            self.row_width)

        rect = mpatches.Rectangle(
            (interval_start + self.peptide_pad, self.cur_y),
            width=(interval_end - interval_start) - self.peptide_end_pad,
            height=self.layer_height,
            facecolor=color, edgecolor='none',
            alpha=alpha)
        self._pack_sequence_metadata(rect, gpm)

        self.ax.add_patch(rect)
        return interval_start, interval_end

    def _compute_sequence_indices(self, gpm, current_position):
        # Compute offsets into the peptide sequence to select
        # PTMs to draw for this row
        if (current_position) > gpm.start_position:
            start_index = current_position - gpm.start_position
            if gpm.end_position - start_index > self.row_width:
                end_index = min(
                    self.row_width,
                    len(self._get_sequence(gpm)))
            else:
                end_index = gpm.end_position - start_index
        else:
            start_index = min(0, gpm.start_position - current_position)
            end_index = min(
                gpm.end_position - current_position,
                self.row_width - (gpm.start_position - current_position))
        return start_index, end_index

    def _compute_sequence_color(self, gpm):
        color = "lightblue"
        alpha = min(max(self.sort_key(gpm) * 2, 0.2), 0.8)
        return color, alpha

    def _compute_modification_color(self, gpm, modification):
        color = get_color(modification.name)
        return color

    def draw_modification_chips(self, gpm, current_position):
        start_index, end_index = self._compute_sequence_indices(gpm, current_position)

        # Extract PTMs from the peptide sequence to draw over the
        # peptide rectangle
        seq = self._get_sequence(gpm)

        for i, pos in enumerate(seq[start_index:end_index]):
            if pos.modifications:
                modification = pos.modifications[0]
                color = self._compute_modification_color(gpm, modification)
                facecolor, edgecolor = lighten(
                    color), darken(color, 0.6)

                mod_patch = mpatches.Rectangle(
                    (gpm.start_position - current_position +
                     i - self.mod_x_offset + 0.3 + start_index, self.cur_y),
                    width=self.mod_width, height=self.layer_height, alpha=0.4,
                    facecolor=facecolor, edgecolor=edgecolor, linewidth=0.5,
                )

                self.id_mapper.add(
                    "modification-%d", mod_patch,
                    {
                        "modification-type": str(modification),
                        "parent": gpm.id
                    })
                self.ax.add_patch(mod_patch)
                modification_string = str(modification)
                modification_symbol = modification_string[0]
                if modification_symbol == '@':
                    modification_symbol = modification_string[1]
                text_path = TextPath(
                    (gpm.start_position - current_position + i -
                     self.mod_text_x_offset + 0.3 + start_index,
                     self.cur_y + self.mod_text_y_offset),
                    modification_symbol, size=self.mod_font_size / 4.5, prop=font_options)
                patch = mpatches.PathPatch(
                    text_path, facecolor='black', lw=0.04)
                self.ax.add_patch(patch)

    def draw_peptidoform(self, gpm, current_position, next_row):
        self.draw_peptide_block(gpm, current_position, next_row)
        self.draw_modification_chips(gpm, current_position)

    def draw_current_row(self, current_position):
        next_row = current_position + self.row_width
        for layer in self.layers:
            c = 0
            for gpm in layer:
                if gpm.start_position < current_position and gpm.end_position < current_position:
                    continue
                elif gpm.start_position >= next_row:
                    break
                c += 1
                self.draw_peptidoform(gpm, current_position, next_row)

            if c > 0:
                self.cur_y += self.y_step

    def finalize_axes(self, ax=None, remove_axes=True):
        if ax is None:
            ax = self.ax
        ax.set_ylim(self.cur_y - 5, 0.2)
        ax.set_xlim(-3., self.row_width + 1)
        if remove_axes:
            ax.axis('off')

    def draw(self):
        self.draw_protein_main_sequence(self.cur_position)
        self.draw_current_row(self.cur_position)
        while self.next_row():
            self.draw_protein_main_sequence(self.cur_position)
            self.draw_current_row(self.cur_position)
        self.finalize_axes()
        return self

    def to_svg(self, scale=1.5, height_padding_scale=1.2):
        ax = self.ax
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        ax.autoscale()

        x_size = sum(map(abs, xlim))
        y_size = sum(map(abs, ylim))

        aspect_ratio = x_size / y_size
        canvas_x = 8.
        canvas_y = canvas_x / aspect_ratio

        fig = ax.get_figure()
        fig.tight_layout(pad=0)
        fig.patch.set_visible(False)
        fig.set_figwidth(canvas_x)
        fig.set_figheight(canvas_y)

        ax.patch.set_visible(False)
        buff = BytesIO()
        fig.savefig(buff, format='svg')
        parser = ET.XMLParser(huge_tree=True)
        root, ids = ET.XMLID(buff.getvalue(), parser=parser)
        root.attrib['class'] = 'plot-glycoforms-svg'
        for id, attributes in self.id_mapper.items():
            element = ids[id]
            element.attrib.update({("data-" + k): str(v)
                                   for k, v in attributes.items()})
            element.attrib['class'] = id.rsplit('-')[0]
        width = float(root.attrib["width"][:-2]) * 1.75
        root.attrib['width'] = '%fpt' % width
        root.attrib.pop("viewBox")

        height = width / (aspect_ratio)

        root.attrib["height"] = "%dpt" % (height * height_padding_scale)
        root.attrib["preserveAspectRatio"] = "xMinYMin meet"
        figure_group = root.find("./{http://www.w3.org/2000/svg}g/[@id='figure_1']")
        figure_group.attrib["transform"] = "scale(%f)" % scale
        svg = ET.tostring(root)
        return svg


class CompressedPileupLayout(GlycoformLayout):

    default_protein_bar_color = 'black'
    n_glycosite_bar_color = 'red'

    def __init__(self, protein, glycopeptides, scale_factor=1.0, ax=None, row_width=50, compression=8, **kwargs):
        super(CompressedPileupLayout, self).__init__(
            protein, glycopeptides, scale_factor, ax, row_width, **kwargs)
        self.compress(compression)

    def compress(self, scale):
        self.layer_height /= scale
        self.y_step /= scale

    def layout_layers(self, matches):
        layers = [[]]
        matches = list(matches)
        matches.sort(key=lambda x: getattr(x, "ms2_score", float('inf')), reverse=True)
        for gpm in matches:
            placed = False
            for layer in layers:
                collision = False
                for member in layer:
                    if span_overlap(gpm, member):
                        collision = True
                        break
                if not collision:
                    layer.append(gpm)
                    placed = True
                    break
            if not placed:
                layers.append([gpm])
        return layers

    def _make_text_scaler(self):
        transform = Affine2D()
        transform.scale(self.row_width / 75., 0.5)
        return transform

    def draw_protein_main_sequence(self, current_position):
        next_row = current_position + self.row_width
        transform = self._make_text_scaler()
        for i, aa in enumerate(self.protein.protein_sequence[current_position:next_row]):
            color = self.n_glycosite_bar_color if (i + current_position) in self.glycosites\
                else self.default_protein_bar_color
            rect = mpatches.Rectangle(
                (self.protein_pad + i, self.layer_height + .05 + self.cur_y),
                width=self.sequence_font_size / 4.5,
                height=self.sequence_font_size / 30.,
                facecolor=color)
            self.ax.add_patch(rect)
            if i % 100 == 0 and i != 0:
                xy = np.array((self.protein_pad + i, self.layer_height + .35 + self.cur_y))
                text_path = TextPath(
                    xy,
                    str(current_position + i), size=self.sequence_font_size / 7.5,
                    prop=font_options)
                text_path = text_path.transformed(transform)
                new_center = transform.transform(xy)
                delta = xy - new_center - (1, 0)
                text_path = text_path.transformed(Affine2D().translate(*delta))
                patch = mpatches.PathPatch(text_path, facecolor='grey', lw=0.04)
                self.ax.add_patch(patch)

    def _pack_sequence_metadata(self, rect, gpm):
        pass

    def _get_sequence(self, gpm):
        try:
            glycopeptide = gpm.structure
            glycopeptide = PeptideSequence(str(glycopeptide))
            return glycopeptide
        except AttributeError:
            return PeptideSequence(str(gpm))

    def _compute_sequence_color(self, gpm):
        try:
            glycopeptide = gpm.structure
            glycopeptide = PeptideSequence(str(glycopeptide))
            if "N-Glycosylation" in glycopeptide.modification_index:
                return 'forestgreen', 0.5
            elif 'O-Glycosylation' in glycopeptide.modification_index:
                return 'aquamarine', 0.5
            elif 'GAG-Linker' in glycopeptide.modification_index:
                return 'orange', 0.5
            else:
                raise ValueError(glycopeptide)
        except AttributeError:
            return 'red', 0.5

    def draw_modification_chips(self, gpm, current_position):
        return

    def finalize_axes(self, ax=None, remove_axes=True):
        super(CompressedPileupLayout, self).finalize_axes(ax, remove_axes)
        self.ax.set_xlim(1., self.row_width + 2)


class DigestLayout(GlycoformLayout):
    def __init__(self, *args, **kwargs):
        super(DigestLayout, self).__init__(*args, **kwargs)
        protease = enzyme.Protease(kwargs.get("enzyme", "trypsin"))
        self.cleavage_sites = set(i for c in protease.cleave(str(self.protein)) for i in c[1:3] if i != 0)

    def _draw_main_sequence(self, start, end):
        for i, aa in enumerate(self.protein.protein_sequence[start:end]):
            text_path = TextPath(
                (self.protein_pad + i, self.layer_height + .2 + self.cur_y),
                aa, size=self.sequence_font_size / 7.5, prop=font_options)
            color = 'red' if any(
                (((i + start) in self.glycosites),
                 ((i + start - 1) in self.glycosites),
                 ((i + start - 2) in self.glycosites))
            ) else 'black'
            patch = mpatches.PathPatch(text_path, facecolor=color, edgecolor=color, lw=0.04)
            self.ax.add_patch(patch)
            if i + start in self.cleavage_sites:
                rect = mpatches.Rectangle(
                    (self.protein_pad + i - 0.33,
                     self.layer_height + self.cur_y - 0.25), 0.15, 1.5, fc='black')
                self.ax.add_patch(rect)


def draw_layers(layers, protein, scale_factor=1.0, ax=None, row_width=50, **kwargs):
    '''
    Render fixed-width stacked peptide identifications across
    a protein. Each shape is rendered with a unique identifier.
    '''
    if ax is None:
        figure, ax = plt.subplots(1, 1)
    id_mapper = IDMapper()
    i = 0

    layer_height = 0.56 * scale_factor
    y_step = (layer_height + 0.15) * -scale_factor
    cur_y = -3

    cur_position = 0

    mod_text_x_offset = 0.50 * scale_factor
    sequence_font_size = 6. * scale_factor
    mod_font_size = 2.08 * scale_factor
    mod_text_y_offset = 0.1 * scale_factor
    mod_width = 0.5 * scale_factor
    mod_x_offset = 0.60 * scale_factor
    total_length = len(protein.protein_sequence or '')
    protein_pad = -0.365 * scale_factor
    peptide_pad = protein_pad * (1.2)
    peptide_end_pad = 0.35 * scale_factor

    glycosites = set(protein.n_glycan_sequon_sites)
    for layer in layers:
        layer.sort(key=lambda x: x.start_position)

    while cur_position < total_length:
        next_row = cur_position + row_width
        i = -2
        text_path = TextPath(
            (protein_pad + i, layer_height + .2 + cur_y),
            str(cur_position + 1), size=sequence_font_size / 7.5, prop=font_options, stretch=1000)
        patch = mpatches.PathPatch(text_path, facecolor='grey', lw=0.04)
        ax.add_patch(patch)

        i = row_width + 2
        text_path = TextPath(
            (protein_pad + i, layer_height + .2 + cur_y),
            str(next_row), size=sequence_font_size / 7.5, prop=font_options, stretch=1000)
        patch = mpatches.PathPatch(text_path, facecolor='grey', lw=0.04)
        ax.add_patch(patch)

        for i, aa in enumerate(protein.protein_sequence[cur_position:next_row]):
            text_path = TextPath(
                (protein_pad + i, layer_height + .2 + cur_y),
                aa, size=sequence_font_size / 7.5, prop=font_options, stretch=1000)
            color = 'red' if any(
                (((i + cur_position) in glycosites),
                 ((i + cur_position - 1) in glycosites),
                 ((i + cur_position - 2) in glycosites))
            ) else 'black'
            patch = mpatches.PathPatch(text_path, facecolor=color, lw=0.04)
            ax.add_patch(patch)

        for layer in layers:
            c = 0
            for gpm in layer:
                if gpm.start_position < cur_position and gpm.end_position < cur_position:
                    continue
                elif gpm.start_position >= next_row:
                    break
                c += 1

                color = "lightblue"
                alpha = min(max(gpm.ms2_score * 2, 0.2), 0.8)

                interval_start = max(
                    gpm.start_position - cur_position,
                    0)
                interval_end = min(
                    len(gpm.structure) + gpm.start_position - cur_position,
                    row_width)

                rect = mpatches.Rectangle(
                    (interval_start + peptide_pad, cur_y),
                    width=(interval_end - interval_start) - peptide_end_pad,
                    height=layer_height,
                    facecolor=color, edgecolor='none',
                    alpha=alpha)

                id_mapper.add("glycopeptide-%d", rect, {
                    "sequence": str(gpm.structure),
                    "start-position": gpm.start_position,
                    "end-position": gpm.end_position,
                    "ms2-score": gpm.ms2_score,
                    "q-value": gpm.q_value,
                    "record-id": gpm.id if hasattr(gpm, 'id') else None,
                    "calculated-mass": gpm.structure.total_mass,
                    "spectra-count": len(gpm.spectrum_matches)
                })
                ax.add_patch(rect)

                # Compute offsets into the peptide sequence to select
                # PTMs to draw for this row
                if (cur_position) > gpm.start_position:
                    start_index = cur_position - gpm.start_position
                    if gpm.end_position - start_index > row_width:
                        end_index = min(
                            row_width,
                            len(gpm.structure))
                    else:
                        end_index = gpm.end_position - start_index
                else:
                    start_index = min(0, gpm.start_position - cur_position)
                    end_index = min(
                        gpm.end_position - cur_position,
                        row_width - (gpm.start_position - cur_position))

                # Extract PTMs from the peptide sequence to draw over the
                # peptide rectangle
                seq = gpm.structure

                for i, pos in enumerate(seq[start_index:end_index]):
                    if len(pos[1]) > 0:
                        color = get_color(pos[1][0].name)
                        facecolor, edgecolor = lighten(
                            color), darken(color, 0.6)

                        mod_patch = mpatches.Rectangle(
                            (gpm.start_position - cur_position +
                             i - mod_x_offset + 0.3 + start_index, cur_y),
                            width=mod_width, height=layer_height, alpha=0.4,
                            facecolor=facecolor, edgecolor=edgecolor, linewidth=0.5,
                        )

                        id_mapper.add(
                            "modification-%d", mod_patch,
                            {
                                "modification-type": pos[1][0].name,
                                "parent": gpm.id
                            })
                        ax.add_patch(mod_patch)
                        text_path = TextPath(
                            (gpm.start_position - cur_position + i -
                             mod_text_x_offset + 0.3 + start_index, cur_y + mod_text_y_offset),
                            str(pos[1][0])[0], size=mod_font_size / 4.5, prop=font_options)
                        patch = mpatches.PathPatch(
                            text_path, facecolor='black', lw=0.04)
                        ax.add_patch(patch)
            if c > 0:
                cur_y += y_step
        cur_y += y_step * 3
        cur_position = next_row

    ax.set_ylim(cur_y - 5, 5)
    ax.set_xlim(-5, row_width + 5)
    ax.axis('off')
    return ax, id_mapper


def plot_glycoforms(protein, identifications, **kwargs):
    layout = GlycoformLayout(protein, identifications, **kwargs)
    layout.draw()
    return layout.ax, layout.id_mapper


def plot_glycoforms_svg(protein, identifications, scale=1.5, ax=None,
                        margin_left=80, margin_top=0, height_padding_scale=1.2,
                        **kwargs):
    '''
    A specialization of :func:`plot_glycoforms` which adds additional features to SVG images, such as
    adding shape metadata to XML tags and properly configuring the viewport and canvas for the figure's
    dimensions.

    TODO: replace uses of this function with :meth:`GlycoformLayout.to_svg`
    '''
    ax, id_mapper = plot_glycoforms(protein, identifications, ax=ax, **kwargs)
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    ax.autoscale()

    x_size = sum(map(abs, xlim))
    y_size = sum(map(abs, ylim))

    aspect_ratio = x_size / y_size
    canvas_x = 8.
    canvas_y = canvas_x / aspect_ratio

    fig = ax.get_figure()
    # fig.tight_layout(pad=0.2)
    fig.tight_layout(pad=0)
    fig.patch.set_visible(False)
    fig.set_figwidth(canvas_x)
    fig.set_figheight(canvas_y)

    ax.patch.set_visible(False)
    buff = BytesIO()
    fig.savefig(buff, format='svg')
    root, ids = ET.XMLID(buff.getvalue())
    root.attrib['class'] = 'plot-glycoforms-svg'
    for id, attributes in id_mapper.items():
        element = ids[id]
        element.attrib.update({("data-" + k): str(v)
                               for k, v in attributes.items()})
        element.attrib['class'] = id.rsplit('-')[0]
    min_x, min_y, max_x, max_y = map(float, root.attrib["viewBox"].split(" "))
    min_x += margin_left
    min_y += margin_top
    max_x += 200
    view_box = ' '.join(map(str, (min_x, min_y, max_x, max_y)))
    root.attrib["viewBox"] = view_box
    width = float(root.attrib["width"][:-2]) * 1.75
    root.attrib["width"] = "100%"

    height = width / (aspect_ratio)

    root.attrib["height"] = "%dpt" % (height * height_padding_scale)
    root.attrib["preserveAspectRatio"] = "xMinYMin meet"
    root[1].attrib["transform"] = "scale(%f)" % scale
    svg = ET.tostring(root)
    plt.close()

    return svg



cimport cython

from cpython.list cimport PyList_Size, PyList_GET_ITEM
from ms_deisotope._c.peak_dependency_network.intervals cimport SpanningMixin


cdef class QueryIntervalBase(SpanningMixin):

    cdef:
        public double center

    def __init__(self, center, start, end):
        self.center = center
        self.start = start
        self.end = end

    def __hash__(self):
        return hash((self.start, self.center, self.end))

    def __eq__(self, other):
        return (self.start, self.center, self.end) == (other.start, other.center, other.end)

    def __repr__(self):
        return "QueryInterval(%0.4f, %0.4f)" % (self.start, self.end)

    def __reduce__(self):
        return QueryIntervalBase, (self.center, self.start, self.end)

    cpdef QueryIntervalBase extend(self, QueryIntervalBase other):
        self.start = min(self.start, other.start)
        self.end = max(self.end, other.end)
        self.center = (self.start + self.end) / 2.
        return self

    @staticmethod
    cdef QueryIntervalBase _create(double center, double start, double end):
        cdef QueryIntervalBase new = QueryIntervalBase.__new__(QueryIntervalBase)
        new.center = center
        new.start = start
        new.end = end
        return new

    cpdef QueryIntervalBase scale(self, double x):
        new = QueryIntervalBase._create(
            self.center,
            self.center - ((self.center - self.start) * x),
            self.center + ((self.end - self.center) * x))
        return new

    cpdef QueryIntervalBase copy(self):
        return QueryIntervalBase._create(self.center, self.start, self.end)

    cpdef QueryIntervalBase difference(self, SpanningMixin other):
        if self.contains_interval(other):
            return self.copy()
        elif self.is_contained_in_interval(other):
            return QueryIntervalBase._create(self.center, self.center, self.center)
        if self.start < other.start:
            if self.end < other.start:
                return self.copy()
            else:
                return QueryIntervalBase._create(self.center, self.start, other.start)
        elif self.start > other.end:
            return self.copy()
        elif self.start <= other.end:
            return QueryIntervalBase._create(self.center, other.end, self.end)
        else:
            return self.copy()


cdef class PPMQueryInterval(QueryIntervalBase):

    def __init__(self, mass, error_tolerance=2e-5):
        self.center = mass
        self.start = mass - (mass * error_tolerance)
        self.end = mass + (mass * error_tolerance)


cdef class FixedQueryInterval(QueryIntervalBase):

    def __init__(self, mass, width=3):
        self.center = mass
        self.start = mass - width
        self.end = mass + width


cdef class IntervalFilter(object):
    cdef:
        public list intervals

    def __init__(self, intervals):
        self.intervals = list(intervals)

    cpdef bint test(self, double mass):
        cdef:
            size_t i, n
            SpanningMixin v
            list intervals
        intervals = self.intervals
        n = PyList_Size(intervals)
        for i in range(n):
            v = <SpanningMixin>PyList_GET_ITEM(intervals, i)
            if v._contains(mass):
                return True
        return False

    def __iter__(self):
        return iter(self.intervals)

    def __getitem__(self, i):
        return self.intervals[i]

    def __len__(self):
        return len(self.intervals)

    def __call__(self, mass):
        return self.test(mass)

    cpdef IntervalFilter compress(self):
        cdef:
            size_t i, n
            list out
            QueryIntervalBase last, interval
        n = PyList_Size(self.intervals)
        out = []
        if n == 0:
            return self
        last = <QueryIntervalBase>PyList_GET_ITEM(self.intervals, 0)
        for i in range(1, n):
            interval = <QueryIntervalBase>PyList_GET_ITEM(self.intervals, i)
            if interval.overlaps(last):
                last.extend(interval)
            else:
                out.append(last)
                last = interval
        out.append(last)
        return self




from glypy import monosaccharides, GlycanComposition
from glypy.structure import (
    Modification, Glycan, Substituent)


def gag_linkers():
    xyl = monosaccharides.Xyl
    gal = monosaccharides.Gal
    xyl.add_monosaccharide(gal, 4, child_position=1)
    gal2 = monosaccharides.Gal
    gal.add_monosaccharide(gal2, 3, child_position=1)
    hexa = monosaccharides.Hex
    hexa.add_modification(Modification.Acidic, 6)
    gal2.add_monosaccharide(hexa, 3, child_position=1)
    hexnac = monosaccharides.HexNAc
    hexa.add_monosaccharide(hexnac, 3, child_position=1)
    enhexa = monosaccharides.Hex
    enhexa.add_modification(Modification.Acidic, 6)
    enhexa.add_modification(Modification.en, 2)
    hexnac.add_monosaccharide(enhexa, 3, child_position=1)
    base_linker = Glycan(xyl).reindex()

    variants = [{"substituents": {
        5: "sulfate"
    }}, {
        "monosaccharides": {
            2: "Fuc"
        }
    }, {
        "monosaccharides": {
            2: "Fuc"
        },
        "substituents": {
            5: "sulfate"
        }
    }, {
        "monosaccharides": {
            2: "Fuc"
        },
        "substituents": {
            5: "sulfate",
            3: "sulfate"
        }
    }, {
        "substituents": {
            1: "phosphate"
        }
    }, {
        "substituents": {
            1: "phosphate",
            5: "sulfate"
        }
    }, {
        "substituents": {
            1: "phosphate",
            3: "sulfate",
            5: "sulfate"
        }
    }, {
        "substituents": {
            5: "sulfate"
        }
    }, {
        "substituents": {
            3: "sulfate",
            5: "sulfate"
        }
    }, {
        "monosaccharides": {
            2: "NeuAc"
        }
    }, {
        "monosaccharides": {
            2: "NeuAc"
        },
        "substituents": {
            5: "sulfate"
        }
    }, {
        "monosaccharides": {
            2: "NeuAc"
        },
        "substituents": {
            5: "sulfate",
            3: "sulfate"
        }
    }]

    linker_variants = [base_linker]

    for variant in variants:
        linker = base_linker.clone()

        substituent_variant = variant.get("substituents", {})
        for position, subst in substituent_variant.items():
            parent = linker.get(position)
            parent.add_substituent(Substituent(subst))

        monosacch_variant = variant.get("monosaccharides", {})
        for position, mono in monosacch_variant.items():
            parent = linker.get(position)
            child = monosaccharides[mono]
            parent.add_monosaccharide(child, -1)

        linker.reindex()
        linker_variants.append(linker)

    return linker_variants


def gag_linker_compositions():
    return [GlycanComposition.from_glycan(linker) for linker in gag_linkers()]



from numpy cimport npy_uint32 as uint32_t, npy_uint64 as uint64_t


cdef class GlycopeptideDatabaseRecord(object):
    cdef:
        public size_t id
        public double calculated_mass
        public str glycopeptide_sequence
        public int protein_id
        public int start_position
        public int end_position
        public double peptide_mass
        public int hypothesis_id


cdef class glycopeptide_key(object):
    cdef:
        public uint32_t start_position
        public uint32_t end_position
        public uint64_t peptide_id
        public uint32_t protein_id
        public uint32_t hypothesis_id
        public uint64_t glycan_combination_id
        public object structure_type
        public uint64_t site_combination_index
        public Py_hash_t _hash

    @staticmethod
    cdef glycopeptide_key _create(uint32_t start_position, uint32_t end_position, uint64_t peptide_id,
                                  uint32_t protein_id, uint32_t hypothesis_id, uint64_t glycan_combination_id,
                                  object structure_type, uint64_t site_combination_index, Py_hash_t _hash=*)

    cdef void _rehash(self)
    cpdef glycopeptide_key copy(self, object structure_type=*)
    cpdef dict as_dict(self, bint stringify=*)


cdef class PeptideDatabaseRecordBase(object):
    cdef:
        public size_t id
        public double calculated_mass
        public basestring modified_peptide_sequence
        public uint32_t protein_id
        public int start_position
        public int end_position
        public uint32_t hypothesis_id
        public tuple n_glycosylation_sites
        public tuple o_glycosylation_sites
        public tuple gagylation_sites

    cpdef bint has_glycosylation_sites(self)


cdef class GlycopeptideFragmentCachingContext(object):
    cdef:
        public dict store

    cpdef peptide_backbone_fragment_key(self, target, args, dict kwargs)
    cpdef stub_fragment_key(self, target, args, dict kwargs)
    cpdef bind(self, target)
    cpdef unbind(self, target)
    cpdef _make_target_key(self, tuple key)


cdef class GlycanAwareGlycopeptideFragmentCachingContext(GlycopeptideFragmentCachingContext):
    pass


cpdef list clone_stub_fragments(list stubs)